"""Process the raw ASED dataset.

This assumes the file structure from the original dataset:
/.../
    01Neutral/
        *.wav
    02Fearful/
        *.wav
"""

import shutil
from pathlib import Path

import click
from joblib import delayed

from ertk.dataset import write_annotations, write_filelist
from ertk.utils import TqdmParallel

emotion_map = {
    "a": "anger",
    "h": "happiness",
    "n": "neutral",
    "f": "fear",
    "s": "sadness",
}

# These come from the Microsoft Speech API with their Amharic model.
# I am waiting for a response from the ASED authors for the actual
# sentences.
sentence_map = {
    "n01": "እህቴ እኮ በአውሮፕላን የመጣች ነው",
    "n02": "ሰላም በየቀኑ ልብስ ማጠብ ትወዳለች",
    "n03": "ነገ ስብሰባ አለ",
    "n04": "ሰኔ 30 ትምህርት ይዘጋል",
    "n05": "በቅርቡ ይመጣሉ",
    "n06": "ደህና እደሩ ነገም እንገናኝ",
    "n07": "ቡና ይጠጣል ጠብቀኝ",
    "f01": "እኔ እኮ በአውሮፕላን የመጣች ነው",
    "f02": "ሰላም በየቀኑ ልብስ ማጠብ ትወዳለች",
    "f03": "አንቺ አለቆች መምጣት ሁሉ ግቢ",
    "f04": "አረብ በሩ ተንኳኳ ልየዋ ነው መሰለኝ",
    "f05": "አንተ ቀስ ብለህ አውራ ይሰማዋል",
    "f06": "እሹ ተጋጨ የማያስከድኑ",
    "f07": "የዘመኑ ቤት ተቃጠለ",
    "h01": "እህቴ እኮ በአውሮፕላን የመጣች ነው",
    "h02": "ሰላም በየቀኑ ልብስ ማጠብ ትወዳለች",
    "h03": "አባቴ ኮ ስልክ ገዛ",
    "h04": "ዛሬኮ ልደት",
    "h05": "የብሔር በጣም ደስ ይላል",
    "h06": "አስቴር ወንድ ልጅ ወለደች",
    "h07": "ታሪኩ አዲስ ዓመት ነው",
    "s01": "እህቴ እኮ በአውሮፕላን የመጣች ነው",
    "s02": "ሰላም በየቀኑ ልብስ ማጠብ ትወዳለች",
    "s03": "የአክስት ልጅ በመኪና አደጋ ሞተ ምንጭ",
    "s04": "አጎቴ እኮ ታሞ ሐኪም ቤት ተገኘ",
    "s05": "ወያኔ ሰው ሁሉ በጦርነት አለቀ",
    "s06": "ዓይን መኪና ይዞ ተገለበጠ",
    "s07": "ወያኔ እኮ ትዝ አረፈች",
    "a01": "እህቴ እኮ በአውሮፕላን የመጣች ነው",
    "a02": "ሰላም በየቀኑ ልብስ ማጠብ ትወዳለች",
    "a03": "የራሱ ጉዳይ ለክልል",
    "a04": "በሰው ጉዳይ ጣልቃ አትግቡ",
    "a05": "እዚህ ምን ታደርጋለህ ሥራና አሰራር",
    "a06": "ከዚህ በፊት ቦታ",
    "a07": "ጦር በእጥፍ አለኝ ከአሁን በኋላ እንጃ ላይ",
}


@click.command()
@click.argument(
    "input_dir", type=click.Path(exists=True, file_okay=False, path_type=Path)
)
@click.option("--resample/--noresample", default=True)
def main(input_dir: Path, resample: bool):
    """Process the ASED dataset at location INPUT_DIR and resample audio
    to 16 kHz 16-bit WAV audio.
    """

    paths = list(input_dir.glob("**/*.wav"))
    resample_dir = Path("resampled")
    if resample:
        resample_dir.mkdir(exist_ok=True, parents=True)
        # S4-04-02-01-29 has an upper-case S instead of lower case
        TqdmParallel(len(paths), "Copying files")(
            delayed(shutil.copyfile)(path, resample_dir / (path.stem.lower() + ".wav"))
            for path in paths
        )
    paths = list(resample_dir.glob("*.wav"))
    write_filelist(paths, "files_all")

    transcripts = {}
    for path in paths:
        sentence = path.stem[3:5]
        emotion = path.stem[0].lower()
        transcripts[path.stem] = sentence_map[f"{emotion}{sentence}"]
    write_annotations(transcripts, "transcript")

    write_annotations({p.stem: emotion_map[p.stem[0].lower()] for p in paths}, "label")
    write_annotations(
        {p.stem: ["F", "M"][int(p.stem[9:11]) - 1] for p in paths}, "gender"
    )
    write_annotations({p.stem: p.stem[3:5] for p in paths}, "sentence")
    write_annotations({p.stem: p.stem[6:8] for p in paths}, "repetition")
    write_annotations({p.stem: p.stem[-2:] for p in paths}, "speaker")
    write_annotations({p.stem: "am" for p in paths}, "language")


if __name__ == "__main__":
    main()



"""Utilities for scikit-learn estimators."""

import warnings
from typing import List

import numpy as np
from joblib import Parallel, delayed
from sklearn.base import BaseEstimator, MetaEstimatorMixin, clone
from sklearn.model_selection import GridSearchCV
from sklearn.multiclass import OneVsRestClassifier as _SKOvR
from sklearn.multiclass import _ConstantPredictor
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import LabelBinarizer

__all__ = [
    "get_estimator_tree",
    "get_base_estimator",
    "OneVsRestClassifier",
    "GridSearchVal",
]


# Workaround for OneVsRest
# https://github.com/scikit-learn/scikit-learn/issues/10882#issuecomment-376912896
# https://stackoverflow.com/a/49535681/10044861
def _fit_binary(estimator, X, y, classes=None, **kwargs):
    unique_y = np.unique(y)
    if len(unique_y) == 1:
        if classes is not None:
            if y[0] == -1:
                c = 0
            else:
                c = y[0]
            warnings.warn(
                "Label %s is present in all training examples." % str(classes[c])
            )
        estimator = _ConstantPredictor().fit(X, unique_y)
    else:
        estimator = clone(estimator)
        estimator.fit(X, y, **kwargs)
    return estimator


class OneVsRestClassifier(_SKOvR):
    def fit(self, X, y, **kwargs):
        self.label_binarizer_ = LabelBinarizer(sparse_output=True)
        Y = self.label_binarizer_.fit_transform(y)
        Y = Y.tocsc()
        self.classes_ = self.label_binarizer_.classes_
        columns = (col.toarray().ravel() for col in Y.T)
        self.estimators_ = Parallel(n_jobs=self.n_jobs)(
            delayed(_fit_binary)(
                self.estimator,
                X,
                column,
                classes=[
                    "not %s" % self.label_binarizer_.classes_[i],
                    self.label_binarizer_.classes_[i],
                ],
                **kwargs,
            )
            for i, column in enumerate(columns)
        )
        return self


def get_estimator_tree(clf) -> List:
    """Gets the estimator list of nested estimators.

    Parameters
    ----------
    clf: estimator
        An estimator.

    Returns
    -------
    list
        The tree of estimators of `clf`.
    """
    tree = [clf]
    if isinstance(clf, Pipeline):
        for x in clf.steps[:-1]:
            tree += get_estimator_tree(x[1])
        if clf._final_estimator != "passthrough":
            return tree + get_estimator_tree(clf._final_estimator)
    for attr in ["base_estimator", "estimator"]:
        if hasattr(clf, attr):
            return tree + get_estimator_tree(getattr(clf, attr))
    return tree


def get_base_estimator(clf) -> BaseEstimator:
    """Gets the base estimator of a pipeline or meta-estimator, assuming
    there is only one.

    Parameters
    ----------
    clf: estimator
        An estimator.

    Returns
    -------
    estimator
        The base estimator of `clf`.
    """
    if isinstance(clf, Pipeline):
        if clf._final_estimator == "passthrough":
            raise ValueError("Pipeline final estimator is 'passthrough'.")
        return get_base_estimator(clf._final_estimator)
    if isinstance(clf, MetaEstimatorMixin):
        for attr in ["base_estimator", "estimator"]:
            if hasattr(clf, attr):
                return get_base_estimator(getattr(clf, attr))
        raise RuntimeError("Couldn't get base estimator from meta estimator")
    return clf


class GridSearchVal(GridSearchCV):
    """Grid-search but using a single validation set. Simple hack of
    GridSearchCV to avoid retraining on train + val data.
    """

    def __init__(
        self,
        estimator,
        param_grid,
        *,
        scoring=None,
        n_jobs=None,
        refit=False,
        cv=None,
        verbose=0,
        pre_dispatch="2*n_jobs",
        error_score=np.nan,
        return_train_score=False,
    ):
        super().__init__(
            estimator,
            param_grid,
            scoring=scoring,
            n_jobs=n_jobs,
            refit=False,
            cv=cv,
            verbose=verbose,
            pre_dispatch=pre_dispatch,
            error_score=error_score,
            return_train_score=return_train_score,
        )

    def fit(self, X, y=None, *, groups=None, **fit_params):
        from sklearn.utils.validation import _check_fit_params

        self.refit = False
        super().fit(X, y=y, groups=groups, **fit_params)

        # Refit only on real train part of "train" data
        train, _ = next(iter(self.cv.split(X, y, groups)))
        fit_params = _check_fit_params(X, fit_params, indices=train)
        self.best_estimator_ = clone(
            clone(self.estimator).set_params(**self.best_params_)
        )
        self.best_estimator_.fit(X[train], y[train], **fit_params)
        return self.best_estimator_



import warnings
from abc import ABC
from dataclasses import dataclass, field
from typing import Any, ClassVar, Dict, List, Optional, Type, cast

import yaml
from omegaconf import MISSING, OmegaConf
from sklearn.ensemble import RandomForestClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.model_selection import GridSearchCV, ParameterGrid
from sklearn.neighbors import KNeighborsClassifier
from sklearn.neural_network import MLPClassifier
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import MaxAbsScaler, MinMaxScaler, Normalizer, StandardScaler

from ertk.config import ERTKConfig

CLASSIFIER_MAP = {
    "rf": RandomForestClassifier,
    "knn": KNeighborsClassifier,
    "lr": LogisticRegression,
    "mlp": MLPClassifier,
}

TRANSFORM_MAP = {
    "std": StandardScaler,
    "minmax": MinMaxScaler,
    "norm": Normalizer,
    "maxabs": MaxAbsScaler,
}


@dataclass
class SkModelConfig(ERTKConfig):
    """Configuration for a scikit-learn model."""

    kwargs: Dict[str, Any] = field(default_factory=dict)
    """Keyword arguments to pass to the model constructor."""
    transform: Optional[str] = "std"
    """The name of the transform to use."""
    transform_kwargs: Dict[str, Any] = field(default_factory=dict)
    """Keyword arguments to pass to the transform constructor."""
    param_grid: Dict[str, Any] = field(default_factory=dict)
    """The parameter grid to use for grid search."""
    param_grid_path: Optional[str] = None
    """The path to a YAML file containing the parameter grid to use for
    grid search.
    """


class ERTKSkModel(ABC):
    """Base class for scikit-learn models.

    Parameters
    ----------
    config: SkModelConfig
        The model configuration.
    """

    config: SkModelConfig
    """The model configuration."""

    _config_type: ClassVar[Type[SkModelConfig]]
    _friendly_name: ClassVar[str]
    _registry: ClassVar[Dict[str, Type["ERTKSkModel"]]] = {}

    def __init__(self, config: SkModelConfig) -> None:
        full_config = OmegaConf.merge(type(self).get_default_config(), config)
        # Inplace merge so that subclass __init__() also gets the full config
        config.merge_with(full_config)  # type: ignore

        self.config = config

    def __init_subclass__(
        cls, fname: str = None, config: Type[SkModelConfig] = None
    ) -> None:
        cls._registry = {}
        if fname and config:
            cls._friendly_name = fname
            cls._config_type = config
            for t in cls.mro()[1:-1]:
                t = cast(Type[ERTKSkModel], t)  # For MyPy
                if not hasattr(t, "_registry"):
                    continue
                if fname in t._registry:
                    prev_cls = t._registry[fname]
                    msg = f"Name {fname} already registered with class {prev_cls}."
                    if prev_cls is cls:
                        warnings.warn(msg)
                    else:
                        raise KeyError(msg)
                t._registry[fname] = cls

    @classmethod
    def get_model_class(cls, name: str) -> Type["ERTKSkModel"]:
        try:
            return cls._registry[name]
        except KeyError as e:
            raise ValueError(f"No model named {name}") from e

    @classmethod
    def make_model(cls, name: str, config: SkModelConfig) -> "ERTKSkModel":
        return cls.get_model_class(name)(config)

    @classmethod
    def get_config_type(cls) -> Type[SkModelConfig]:
        return cls._config_type

    @classmethod
    def valid_models(cls) -> List[str]:
        return list(cls._registry)

    @classmethod
    def get_default_config(cls) -> SkModelConfig:
        return OmegaConf.structured(cls._config_type)


class SkWrapperConfig(SkModelConfig):
    """Configuration for a scikit-learn model wrapper."""

    kind: str = MISSING
    """The name of the model to wrap."""


class SkWrapperModel(ERTKSkModel, fname="wrapper", config=SkWrapperConfig):
    """A wrapper for scikit-learn models."""

    config: SkWrapperConfig

    def __init__(self, config: SkWrapperConfig) -> None:
        super().__init__(config)

        self.kind = config.kind
        clf = CLASSIFIER_MAP[self.kind](**config.kwargs)

        transform = None
        if config.transform:
            transform = TRANSFORM_MAP[config.transform](**config.transform_kwargs)
        clf = Pipeline([("transform", transform), ("clf", clf)])

        param_grid = {}
        if config.param_grid_path:
            with open(config.param_grid_path) as fid:
                param_grid = yaml.safe_load(fid)
        elif config.param_grid:
            param_grid = config.param_grid
        if len(ParameterGrid(param_grid)) > 1:
            clf = GridSearchCV(clf, param_grid)

        self.clf = clf


def gen_wrapper_for_clf(kind: str) -> None:
    """Generate a wrapper for a scikit-learn classifier.

    Parameters
    ----------
    kind: str
        The name of the classifier.
    """

    class _(ERTKSkModel, fname=kind, config=SkModelConfig):
        def __init__(self, config: SkModelConfig) -> None:
            super().__init__(config)
            self.kind = kind
            clf = CLASSIFIER_MAP[kind](**config.kwargs)

            transform = None
            if config.transform:
                transform = TRANSFORM_MAP[config.transform](**config.transform_kwargs)
            clf = Pipeline([("transform", transform), ("clf", clf)])

            param_grid = {}
            if config.param_grid_path:
                with open(config.param_grid_path) as fid:
                    param_grid = yaml.safe_load(fid)
            elif config.param_grid:
                param_grid = config.param_grid
            if len(ParameterGrid(param_grid)) > 1:
                clf = GridSearchCV(clf, param_grid)

            self.clf = clf


for key in CLASSIFIER_MAP:
    gen_wrapper_for_clf(key)



"""Training configuration for PyTorch models."""

from dataclasses import dataclass, field
from typing import Any, Dict, Optional

import omegaconf

from ertk.config import ERTKConfig
from ertk.pytorch.models import PyTorchModelConfig

__all__ = [
    "PyTorchTrainConfig",
    "PyTorchLoggingConfig",
    "PyTorchDataConfig",
    "PyTorchDataAugConfig",
]


@dataclass
class PyTorchLoggingConfig(ERTKConfig):
    log_dir: Optional[str] = ""
    tensorboard: bool = True
    csv: bool = True


@dataclass
class PyTorchDataAugConfig(ERTKConfig):
    spec: str = omegaconf.MISSING
    kwargs: Dict[str, Any] = field(default_factory=dict)


@dataclass
class PyTorchDataConfig(ERTKConfig):
    train_data_processing: Dict[str, PyTorchDataAugConfig] = field(default_factory=dict)
    valid_data_processing: Dict[str, PyTorchDataAugConfig] = field(default_factory=dict)
    test_data_processing: Dict[str, PyTorchDataAugConfig] = field(default_factory=dict)


@dataclass
class PyTorchTrainConfig(ERTKConfig):
    batch_size: int = 32
    logging: PyTorchLoggingConfig = PyTorchLoggingConfig()
    n_gpus: int = 1
    dist_strategy: str = "dp"
    epochs: int = 50
    wrapper_model_config: PyTorchModelConfig = PyTorchModelConfig()
    enable_checkpointing: bool = False
    resume_checkpoint: str = ""
    save_every_n_epochs: int = 1
    data_config: PyTorchDataConfig = PyTorchDataConfig()



"""Implementation of Restricted Boltzmann Machines (RBMs) [1]_ in
TensorFlow.

References
----------
.. [1] G. E. Hinton, “A Practical Guide to Training Restricted Boltzmann
       Machines,” in Neural Networks: Tricks of the Trade: Second
       Edition, G. Montavon, G. B. Orr, and K.-R. Müller, Eds., in
       Lecture Notes in Computer Science. Berlin, Heidelberg: Springer
       Berlin Heidelberg, 2012, pp. 599-619. doi:
       10.1007/978-3-642-35289-8_32.
"""

from enum import Enum
from math import pi
from pathlib import Path
from typing import Optional, Sequence, Tuple

import click
import numpy as np
import tensorflow as tf

from ertk.dataset import Dataset, write_features
from ertk.utils import PathOrStr

__all__ = ["BBRBM", "GBRBM"]


def _sample_bernoulli(p: tf.Tensor):
    return tf.nn.relu(tf.sign(p - tf.random.uniform(tf.shape(p))))


def _sample_std_normal(p: tf.Tensor):
    return tf.random.normal()


class DecayType(Enum):
    STEP = "STEP"
    COSINE = "COSINE"
    EXP = "EXP"

    def __str__(self):
        return self.name


class RBM:
    """Restricted Boltzmann Machine

    This implementation is designed around the work of Hinton et. al.
    [1], [2].

    References
    ----------
    .. [1] G. E. Hinton and R. R. Salakhutdinov, 'Reducing the
    Dimensionality of Data with Neural Networks', Science, vol. 313, no.
    5786, pp. 504-507, Jul. 2006, doi: 10.1126/science.1127647.

    .. [2] G. E. Hinton, 'A Practical Guide to Training Restricted
    Boltzmann Machines', in Neural Networks: Tricks of the Trade: Second
    Edition, G. Montavon, G. B. Orr, and K.-R. Müller, Eds. Berlin,
    Heidelberg: Springer Berlin Heidelberg, 2012, pp. 599-619.
    """

    def __init__(
        self,
        n_hidden: int,
        input_shape: Tuple[int, ...],
        logdir: PathOrStr = "logs/rbm",
        output_histograms: bool = False,
    ):
        """
        Parameters
        ----------
        n_hidden: int, default = 128
            Number of hidden units
        input_shape: tuple
            Shape of input tensors, either 1D or 2D
        logdir: str or Path, optional
            Path to store TensorBoard logs. Default is 'logs/rbm'.
        output_histograms: bool, optional
            Whether to output histogram information. Can slow down training
            somewhat. Default is False.
        """
        if any(x is None or x < 0 for x in input_shape):
            raise ValueError("input_shape must be fully specified.")

        self.input_shape = tf.TensorShape(input_shape)
        n_visible = tf.reduce_prod(self.input_shape)

        self.n_hidden = n_hidden

        self.h_bias = tf.Variable(tf.zeros((n_hidden,)))
        self.v_bias = tf.Variable(tf.zeros((n_visible,)))
        self.W = tf.Variable(tf.random.normal((n_visible, n_hidden), 0.0, 0.01))

        self.h_bias_velocity = tf.Variable(tf.zeros_like(self.h_bias), trainable=False)
        self.v_bias_velocity = tf.Variable(tf.zeros_like(self.v_bias), trainable=False)
        self.W_velocity = tf.Variable(tf.zeros_like(self.W), trainable=False)

        self.learning_rate = tf.Variable(0.01, trainable=False)
        self.momentum = tf.Variable(0.5, trainable=False)
        self.l2_weight = tf.Variable(0.001, trainable=False)

        self.output_histograms = output_histograms

        self.logging = False
        if logdir:
            self.logging = True
            logdir = Path(logdir)
            self.train_summary_writer = tf.summary.create_file_writer(
                str(logdir / "train")
            )
            self.valid_summary_writer = tf.summary.create_file_writer(
                str(logdir / "valid")
            )

    def free_energy(self, batch: tf.Tensor):
        raise NotImplementedError("free_energy() not implemented.")

    def hidden_prob(self, batch: tf.Tensor):
        raise NotImplementedError("hidden_prob() not implemented.")

    def visible_prob(self, h: tf.Tensor):
        raise NotImplementedError("visible_prob() not implemented.")

    def sample_h(self, p: tf.Tensor):
        raise NotImplementedError("sample_h() not implemented.")

    def sample_v(self, p: tf.Tensor):
        raise NotImplementedError("sample_v() not implemented.")

    def calculate_deltas(self, batch: tf.Tensor):
        raise NotImplementedError("calculate_deltas() not implemented.")

    def forward(self, batch: tf.Tensor):
        """Samples hidden units from input batch.

        Parameters
        ----------
        batch: tensor of shape (batch, n_visible)
            Batch of tensors to process.

        Returns
        -------
        Sample hidden units.
        """
        p_h = self.hidden_prob(batch)
        h = self.sample_h(p_h)
        return h

    def backward(self, h: tf.Tensor):
        p_v = self.visible_prob(h)
        return p_v

    @tf.function
    def train_batch(self, batch: tf.Tensor):
        """Runs one step of Gibbs sampling and gradient ascent with the given
        batch.

        Parameters
        ----------
        batch: tensor of shape (batch, n_visible)
            Batch of tensors to train on.

        Returns
        -------
        Mean square error (MSE) for this batch.
        """
        p_h = self.hidden_prob(batch)
        h = self.sample_h(p_h)
        p_v = self.visible_prob(h)
        p_hd = tf.sigmoid(self.h_bias + tf.matmul(p_v, self.W))

        # v*h.T - v'*h'.T
        pos = tf.matmul(tf.expand_dims(batch, -1), tf.expand_dims(p_h, -2))
        neg = tf.matmul(tf.expand_dims(p_v, -1), tf.expand_dims(p_hd, -2))
        outers = pos - neg
        W_delta = tf.reduce_mean(outers, 0) - self.l2_weight * 2.0 * self.W
        # h - h'
        h_bias_delta = tf.reduce_mean(p_h - p_hd, 0)
        # v - v'
        v_bias_delta = tf.reduce_mean(batch - p_v, 0)

        self.W_velocity.assign(
            self.momentum * self.W_velocity + self.learning_rate * W_delta
        )
        self.h_bias_velocity.assign(
            self.momentum * self.h_bias_velocity + self.learning_rate * h_bias_delta
        )
        self.v_bias_velocity.assign(
            self.momentum * self.v_bias_velocity + self.learning_rate * v_bias_delta
        )

        self.W.assign_add(self.W_velocity)
        self.h_bias.assign_add(self.h_bias_velocity)
        self.v_bias.assign_add(self.v_bias_velocity)

        mse_batch = tf.reduce_mean((p_v - batch) ** 2)
        return mse_batch

    def train(
        self,
        train_data: tf.data.Dataset,
        valid_data: Optional[tf.data.Dataset] = None,
        init_learning_rate: float = 0.01,
        final_learning_rate: float = 0.001,
        learning_rate_decay: Optional[DecayType] = None,
        learning_rate_decay_param: float = 0.5,
        init_momentum: float = 0.5,
        final_momentum: float = 0.9,
        momentum_decay: Optional[DecayType] = None,
        momentum_decay_param: float = 0.5,
        n_epochs: int = 100,
    ):
        """Train this RBM.

        Parameters
        ----------
        train_data: tf.data.Dataset
            Training data to use.
        valid_data: tf.data.Dataset, optional
            Validation data to use for performance measuring.
        init_learning_rate: float
            The initial learning rate to set. Default is 0.01.
        final_learning_rate: float
            The final learning rate to set. Default is 0.001
        learning_rate_decay: DecayType, optional
            Type of learning rate decay. Default is None, meaning no decay.
        learning_rate_decay_param: float
            Parameter for learning rate decay. Use depends on type. For step
            decay, this is the fraction of epochs after which to change
            learning rate.
        init_momentum: float
            The initial momentum to set. Default is 0.5.
        final_momentum: float
            The final momentum to set. Default is 0.9.
        momentum_decay: DecayType, optional
            Type of momentum decay. Default is None, meaning no decay.
        n_epochs: int
            The number of epochs to train for. A single epoch is one pass over
            the training data. Default is 100.
        """
        self.learning_rate.assign(init_learning_rate)
        self.momentum.assign(init_momentum)
        train_size = sum(1 for _ in train_data)

        if learning_rate_decay == DecayType.COSINE:
            lr_coeffA = 0.5 * (init_learning_rate - final_learning_rate)
            lr_coeffB = 0.5 * (init_learning_rate + final_learning_rate)
        elif learning_rate_decay == DecayType.EXP:
            lr_coeffA = init_learning_rate
            lr_coeffB = tf.math.log(init_learning_rate) - tf.math.log(
                final_learning_rate
            )
        if momentum_decay == DecayType.COSINE:
            mom_coeffA = 0.5 * (init_momentum - final_momentum)
            mom_coeffB = 0.5 * (init_momentum + final_momentum)
        elif momentum_decay == DecayType.EXP:
            mom_coeffA = init_momentum
            mom_coeffB = tf.math.log(init_momentum) - tf.math.log(final_momentum)

        for epoch in range(1, n_epochs + 1):
            tf.print(f"Epoch: {epoch}")
            time_frac = float(epoch) / float(n_epochs)

            if (
                learning_rate_decay == DecayType.STEP
                and time_frac > learning_rate_decay_param
            ):
                self.learning_rate.assign(final_learning_rate)
            elif learning_rate_decay == DecayType.COSINE:
                self.learning_rate.assign(
                    lr_coeffA * tf.cos(pi * time_frac) + lr_coeffB
                )
            elif learning_rate_decay == DecayType.EXP:
                self.learning_rate.assign(lr_coeffA * tf.exp(-lr_coeffB * time_frac))

            if momentum_decay == DecayType.STEP and time_frac > momentum_decay_param:
                self.momentum.assign(final_momentum)
            elif momentum_decay == DecayType.COSINE:
                self.momentum.assign(mom_coeffA * tf.cos(pi * time_frac) + mom_coeffB)
            elif momentum_decay == DecayType.EXP:
                self.momentum.assign(mom_coeffA * tf.exp(-mom_coeffB * time_frac))

            mse_train = tf.TensorArray(tf.float32, size=0, dynamic_size=True)
            i = tf.constant(0, tf.int32)
            for batch in train_data:
                if i % (train_size // 10) == 0:
                    tf.print(".", end="")

                mse_batch = self.train_batch(batch)
                mse_train = mse_train.write(i, mse_batch)
                i += 1

            mse_train = mse_train.stack()
            mse_train = tf.reduce_mean(mse_train)

            with self.train_summary_writer.as_default():
                tf.summary.scalar("mse", mse_train, step=epoch)
                tf.print(f"mse: {mse_train:.3f}, ", end="")

                tf.summary.scalar("learining_rate", self.learning_rate, step=epoch)
                tf.summary.scalar("momentum", self.momentum, step=epoch)
                weights_l2 = tf.reduce_sum(self.W**2)
                tf.summary.scalar("weights_l2", weights_l2, step=epoch)

                if self.output_histograms:
                    tf.summary.histogram("h_bias", self.h_bias, step=epoch)
                    tf.summary.histogram("v_bias", self.v_bias, step=epoch)
                    tf.summary.histogram("weights", self.W, step=epoch)

                    tf.summary.histogram(
                        "h_bias_velocity", self.h_bias_velocity, step=epoch
                    )
                    tf.summary.histogram(
                        "v_bias_velocity", self.v_bias_velocity, step=epoch
                    )
                    tf.summary.histogram(
                        "weights_velocity", self.W_velocity, step=epoch
                    )

                batch = next(iter(train_data))
                f_energy = self.free_energy(batch)
                tf.summary.scalar("free_energy", f_energy, step=epoch)

                if 2 <= len(self.input_shape) <= 3:
                    if len(self.input_shape) == 3:
                        new_shape = tf.concat(
                            ([self.n_hidden], self.input_shape), axis=0
                        )
                    elif len(self.input_shape) == 2:
                        new_shape = tf.concat(
                            ([self.n_hidden], self.input_shape, [1]), axis=0
                        )
                    W = tf.transpose(self.W)
                    W = tf.reshape(W, new_shape)
                    W = (W + 2.0) / 4.0
                    tf.summary.image("hidden_vis", W, step=epoch, max_outputs=128)

                    if len(self.input_shape) == 3:
                        new_shape = tf.concat(([-1], self.input_shape), axis=0)
                    elif len(self.input_shape) == 2:
                        new_shape = tf.concat(([-1], self.input_shape, [1]), axis=0)
                    reconst = self.reconstruct_batch(batch)
                    reconst = tf.reshape(reconst, new_shape)
                    image = tf.reshape(batch, new_shape)
                    combined = tf.concat([image, reconst], 2)
                    tf.summary.image(
                        "reconstruction", combined, step=epoch, max_outputs=10
                    )

            if valid_data:
                mse_valid = tf.TensorArray(tf.float32, size=0, dynamic_size=True)
                i = tf.constant(0, tf.int32)
                for batch in valid_data:
                    p_v = self.reconstruct_batch(batch)
                    mse_batch = tf.reduce_mean((p_v - batch) ** 2)
                    mse_valid = mse_valid.write(i, mse_batch)
                    i += 1
                mse_valid = mse_valid.stack()
                mse_valid = tf.reduce_mean(mse_valid)

                with self.valid_summary_writer.as_default():
                    tf.summary.scalar("mse", mse_valid, step=epoch)
                    tf.print(f"val_mse: {mse_valid:.3f}")

                    batch = next(iter(valid_data))

                    if 2 <= len(self.input_shape) <= 3:
                        if len(self.input_shape) == 3:
                            new_shape = tf.concat(([-1], self.input_shape), axis=0)
                        elif len(self.input_shape) == 2:
                            new_shape = tf.concat(([-1], self.input_shape, [1]), axis=0)
                        reconst = self.reconstruct_batch(batch)
                        reconst = tf.reshape(reconst, new_shape)
                        image = tf.reshape(batch, new_shape)
                        combined = tf.concat([image, reconst], 2)
                        tf.summary.image(
                            "reconstruction", combined, step=epoch, max_outputs=10
                        )

                    p_h = self.hidden_prob(batch)
                    p_h = tf.reshape(p_h, (1, -1, self.n_hidden, 1))
                    p_h = tf.transpose(p_h, [0, 2, 1, 3])
                    tf.summary.image("hidden_prob", p_h, step=epoch)

                    f_energy = self.free_energy(batch)
                    tf.summary.scalar("free_energy", f_energy, step=epoch)

    @tf.function
    def reconstruct_batch(self, batch):
        """Reconstruct a single batch of input.

        Parameters
        ----------
        batch: tensor of shape (batch, n_visible)
            The batch of tensor to reconstruct.

        Returns
        -------
        Reconstructed batch.
        """
        p_h = self.hidden_prob(batch)
        h = self.sample_h(p_h)
        p_v = self.visible_prob(h)
        return p_v

    def reconstruct(self, data: tf.data.Dataset):
        """Reconstruct all input data.

        Parameters
        ----------
        data: tf.data.Dataset
            Data to reconstruct.

        Returns
        -------
        The reconstructed data as a mapped tf.data.Dataset.
        """
        return data.map(self.reconstruct_batch)

    def representations(self, data: tf.data.Dataset):
        return data.map(self.hidden_prob)


class BBRBM(RBM):
    """RBM with Bernoulli visible and hidden units."""

    def free_energy(self, batch: tf.Tensor):
        x = self.h_bias + tf.matmul(batch, self.W)
        f_energy = -tf.reduce_sum(batch * self.v_bias, -1) - tf.reduce_sum(
            tf.nn.softplus(x), -1
        )
        f_energy = tf.reduce_mean(f_energy)
        return f_energy

    def hidden_prob(self, batch: tf.Tensor):
        return tf.sigmoid(self.h_bias + tf.matmul(batch, self.W))

    def visible_prob(self, h: tf.Tensor):
        return tf.sigmoid(self.v_bias + tf.matmul(h, self.W, transpose_b=True))

    def sample_h(self, p: tf.Tensor):
        return _sample_bernoulli(p)


class GBRBM(RBM):
    """RBM with Gaussian visible and Bernoulli hidden units."""

    def free_energy(self, batch: tf.Tensor):
        x = self.h_bias + tf.matmul(batch, self.W)
        f_energy = 0.5 * tf.reduce_sum((batch - self.v_bias) ** 2, -1) - tf.reduce_sum(
            tf.nn.softplus(x), -1
        )
        f_energy = tf.reduce_mean(f_energy)
        return f_energy

    def hidden_prob(self, batch: tf.Tensor):
        return tf.sigmoid(self.h_bias + tf.matmul(batch, self.W))

    def visible_prob(self, h: tf.Tensor):
        return self.v_bias + tf.matmul(h, self.W, transpose_b=True)

    def sample_h(self, p: tf.Tensor):
        return _sample_bernoulli(p)


class DBN:
    """Deep Belief Network, also known as a Deep Boltzmann Machine."""

    def __init__(
        self,
        input_shape: Tuple[int],
        n_layers: int = 3,
        layer_nodes: Sequence[int] = [512, 256, 128],
        logdir: PathOrStr = "logs/dbn",
        **kwargs,
    ):
        """Initialise this DBN.

        Parameters
        ----------
        input_shape: tuple
            Shape of the input tensors to the first layer
        n_layers: int
            Number of hidden layers in this DBN.
        layer_nodes: tuple of int
            Number of hidden nodes in each of the layers. We must have that
            len(layer_nodes) == n_layers
        logdir: str or Path
            Directory to save log files.
        kwargs: dict
            Additional arguments to pass to the RBM constructor.
        """
        if n_layers != len(layer_nodes):
            raise ValueError("layer_nodes must have exactly n_layers elements")

        self.input_shape = input_shape

        logdir = Path(logdir)
        timestamp = tf.timestamp().numpy()
        logdir = logdir / str(timestamp)
        self.layers = [
            BBRBM(layer_nodes[0], self.input_shape, logdir=logdir / "rbm0", **kwargs)
        ]
        for i in range(1, n_layers):
            self.layers.append(
                BBRBM(
                    layer_nodes[i],
                    logdir=logdir / f"rbm{i}",
                    input_shape=(layer_nodes[i - 1],),
                )
            )

        self.train_summary_writer = tf.summary.create_file_writer(
            str(logdir / "dbn" / "train")
        )
        self.valid_summary_writer = tf.summary.create_file_writer(
            str(logdir / "dbn" / "valid")
        )

    def train(
        self,
        train_data: tf.data.Dataset,
        valid_data: Optional[tf.data.Dataset] = None,
        n_epochs: int = 100,
        learning_rate: float = 0.01,
        momentum: float = 0.5,
    ):
        """Train each layer in this DBN in sequence, using the reconstructed
        output from previous layers as input.

        Parameters
        ----------
        train_data: tf.data.Dataset
            Training data to use.
        valid_data: tf.data.Dataset, optional
            Validation data to use for mesuring training performance.
        n_epochs: int
            The number of epochs to train each layer for. Default is 100.
        learning_rate: float
            The learning rate to use for all layers. Default is 0.01.
        momentum: float
            The momentum to use for all layers. Default is 0.5.
        """
        print("Training layer 0")
        self.layers[0].train(
            train_data,
            valid_data=valid_data,
            init_momentum=momentum,
            init_learning_rate=learning_rate,
            n_epochs=n_epochs,
        )
        new_train_data = train_data
        new_valid_data = valid_data
        for i in range(1, len(self.layers)):
            print(f"Training layer {i}")
            new_train_data = new_train_data.map(self.layers[i - 1].forward)
            if new_valid_data:
                new_valid_data = new_valid_data.map(self.layers[i - 1].forward)
            self.layers[i].train(
                new_train_data,
                valid_data=new_valid_data,
                init_momentum=momentum,
                init_learning_rate=learning_rate,
                n_epochs=n_epochs,
            )

        if valid_data:
            with self.valid_summary_writer.as_default():
                batch = next(iter(valid_data))

                if len(self.input_shape) == 3:
                    new_shape = tf.concat(([-1], self.input_shape), axis=0)
                elif len(self.input_shape) == 2:
                    new_shape = tf.concat(([-1], self.input_shape, [1]), axis=0)
                tf.print("Reconstructing")
                reconst = self.reconstruct_batch(batch)
                reconst = tf.reshape(reconst, new_shape)
                image = tf.reshape(batch, new_shape)
                combined = tf.concat([image, reconst], 2)
                tf.summary.image(
                    "reconstruction", combined, step=n_epochs, max_outputs=10
                )

    @tf.function
    def reconstruct_batch(self, batch: tf.Tensor):
        """Reconstruct a single batch of input.

        Parameters
        ----------
        batch: tensor of shape (batch, n_visible)
            The batch of tensor to reconstruct.

        Returns
        -------
        Reconstructed batch.
        """
        for i in range(len(self.layers)):
            batch = self.layers[i].forward(batch)
        for i in range(len(self.layers) - 1, -1, -1):
            batch = self.layers[i].backward(batch)
        return batch

    def reconstruct(self, data: tf.data.Dataset):
        """Reconstruct all input data.

        Parameters
        ----------
        data: tf.data.Dataset
            Data to reconstruct.

        Returns
        -------
        The reconstructed data as a mapped tf.data.Dataset.
        """
        return data.map(self.reconstruct_batch)


@click.command()
@click.argument("input", type=click.Path(exists=True, dir_okay=False, path_type=Path))
@click.option("--logdir", type=str, default="logs/tf/rbm")
@click.option("--batch_size", type=int, default=16)
@click.option("--valid_fraction", type=float, default=0.1)
@click.option("--epochs", type=int, default=200)
@click.option("--init_learning_rate", type=float, default=0.005)
@click.option("--init_momentum", type=float, default=0.5)
@click.option("--final_learning_rate", type=float, default=0.001)
@click.option("--final_momentum", type=float, default=0.9)
@click.option(
    "--learning_rate_decay",
    type=click.Choice(DecayType.__members__),
    default=DecayType.STEP.name,
)
@click.option(
    "--momentum_decay",
    type=click.Choice(DecayType.__members__),
    default=DecayType.STEP.name,
)
@click.option("--output", type=Path)
def main(
    input: Path,
    logdir: str,
    batch_size: int,
    valid_fraction: float,
    epochs: int,
    init_learning_rate: float,
    init_momentum: float,
    final_learning_rate: float,
    final_momentum: float,
    learning_rate_decay: DecayType,
    momentum_decay: DecayType,
    output: Path,
):
    """Trains RBM on spectrograms in INPUT."""

    learning_rate_decay = DecayType(learning_rate_decay)
    momentum_decay = DecayType(momentum_decay)

    for gpu in tf.config.list_physical_devices("GPU"):
        tf.config.experimental.set_memory_growth(gpu, True)

    dataset = Dataset(input)
    x = dataset.x
    input_shape = tuple(x[0].shape)
    xmin = np.min(x, axis=(1, 2), keepdims=True)
    xmax = np.max(x, axis=(1, 2), keepdims=True)
    x = (x - xmin) / (xmax - xmin)
    spectrograms = np.reshape(x, (dataset.n_instances, -1))

    split = int(spectrograms.shape[0] * valid_fraction)
    train_data = tf.data.Dataset.from_tensor_slices(spectrograms[split:, :])
    train_data = train_data.shuffle(spectrograms.shape[0])
    train_data = train_data.batch(batch_size).prefetch(10)
    valid_data = tf.data.Dataset.from_tensor_slices(spectrograms[:split, :])
    valid_data = valid_data.batch(batch_size).prefetch(10)

    Path(logdir).mkdir(parents=True, exist_ok=True)
    rbm = BBRBM(512, input_shape=input_shape, logdir=logdir, output_histograms=True)
    rbm.train(
        train_data,
        valid_data,
        n_epochs=epochs,
        init_learning_rate=init_learning_rate,
        init_momentum=init_momentum,
        final_learning_rate=final_learning_rate,
        final_momentum=final_momentum,
        learning_rate_decay=learning_rate_decay,
        momentum_decay=momentum_decay,
    )
    if output:
        representations = rbm.representations(
            tf.data.Dataset.from_tensor_slices(spectrograms).batch(8)
        )
        representations = np.concatenate(list(representations.as_numpy_iterator()))
        feature_names = [f"bbrbm{i + 1}" for i in range(representations.shape[-1])]
        write_features(
            output,
            names=dataset.names,
            features=representations,
            corpus=dataset.corpus,
            feature_names=feature_names,
        )


if __name__ == "__main__":
    main()



"""Type definitions."""

from os import PathLike
from typing import Callable, Union

import numpy as np

__all__ = ["PathOrStr", "ScoreFunction"]


PathOrStr = Union[PathLike, str]
ScoreFunction = Callable[[np.ndarray, np.ndarray], float]



"""Process the raw EmoV-DB dataset.

This assumes the file structure from the original sorted data:
/.../
    bea/
        Angry/
            *.wav
        Amused/
            *.wav
        ...
    josh/
        Angry/
            *.wav
        ...
    ...
    cmuarctic.data
"""

import re
from pathlib import Path

import click

from ertk.dataset import resample_rename_clips, write_annotations, write_filelist

emotion_map = {
    "amused": "amusement",
    "anger": "anger",
    "disgust": "disgust",
    "neutral": "neutral",
    "sleepiness": "sleepiness",
}

gender_map = {"bea": "F", "jenie": "F", "josh": "M", "sam": "M"}


@click.command()
@click.argument(
    "input_dir", type=click.Path(exists=True, file_okay=False, path_type=Path)
)
@click.option("--resample/--noresample", default=True)
def main(input_dir: Path, resample: bool):
    """Process the EmoV-DB dataset at location INPUT_DIR and resample
    audio to 16 kHz 16-bit WAV audio.
    """
    paths = list(input_dir.glob("**/*.wav"))
    resample_dir = Path("resampled")
    mapping = {x: resample_dir / f"{x.parts[-3]}_{x.stem.lower()}.wav" for x in paths}
    resample_rename_clips(mapping=mapping)
    paths = list(resample_dir.glob("*.wav"))
    write_filelist(paths, "files_all")

    write_annotations(
        {p.stem: emotion_map[p.stem.split("_")[1]] for p in paths}, "label"
    )
    write_annotations({p.stem: p.stem.split("_")[0] for p in paths}, "speaker")
    write_annotations(
        {p.stem: gender_map[p.stem.split("_")[0]] for p in paths}, "gender"
    )
    write_annotations({p.stem: p.stem[-4:] for p in paths}, "sentence")
    write_annotations({p.stem: "us" for p in paths}, "country")
    write_annotations({p.stem: "en" for p in paths}, "language")

    sent_to_transcript = {}
    with open(input_dir / "cmuarctic.data") as fid:
        for line in fid:
            match = re.match(r"^\( arctic_a([0-9]{4}) \"(.*)\" \)$", line.strip())
            if match:
                sent_to_transcript[match.group(1)] = match.group(2)
    write_annotations(
        {p.stem: sent_to_transcript[p.stem[-4:]] for p in paths}, "transcript"
    )


if __name__ == "__main__":
    main()



import os
import subprocess
import sys
from datetime import datetime
from pathlib import Path
from queue import Queue
from threading import Lock, Thread
from typing import Tuple

import click
from click_option_group import RequiredMutuallyExclusiveOptionGroup, optgroup


@click.command()
@click.argument(
    "input", type=click.Path(exists=True, dir_okay=False, path_type=Path), nargs=-1
)
@click.option(
    "--failed", type=Path, help="Where to log failed commands.", required=True
)
@optgroup.group("Parallel method", cls=RequiredMutuallyExclusiveOptionGroup)
@optgroup.option("--cpus", type=int, help="Number of CPU threads to use")
@optgroup.option("--gpus", help="GPU IDs to run on.")
def main(input: Tuple[Path], cpus: int, gpus: str, failed: Path):
    """Run all commands specified in the INPUT file(s), splitting the
    work across multiple CPU threads or GPUs such that each command runs
    solely on whichever thread/GPU is next available.

    Each GPU still has it's own thread to run processes using
    CUDA_VISIBLE_DEVICES. Each thread reads from a synchronous queue and
    runs the command in a shell.
    """
    if len(input) == 0:
        raise ValueError("No input files specified.")

    q: "Queue[str]" = Queue()
    for file in input:
        with open(file) as fid:
            for line in filter(None, map(str.strip, fid)):
                q.put(line)
    n_commands = q.qsize()
    print(f"Loaded {n_commands} commands")

    file_lock = Lock()
    if not failed.exists():
        failed.parent.mkdir(exist_ok=True, parents=True)
    fail_file = open(failed, "w")
    n_failed = 0

    def worker(t_id: int = None, gpu: int = None):
        env = os.environ.copy()
        env.update({"PYTHONUNBUFFERED": "1"})
        if gpu is not None:
            env.update({"CUDA_VISIBLE_DEVICES": str(gpu)})
        else:  # CPU
            env.update({"OMP_NUM_THREADS": "1"})
        print_prefix = f"[T{t_id}]" if gpu is None else f"[GPU{gpu}]"
        while not q.empty():
            cmd = q.get()
            print(
                datetime.now().isoformat(),
                f"Executing command on {print_prefix}: {cmd}",
                flush=True,
            )
            proc = subprocess.Popen(
                cmd,
                env=env,
                shell=True,
                text=True,
                bufsize=1,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
            )
            if proc.stdout is not None:
                with proc.stdout:
                    for line in iter(proc.stdout.readline, ""):
                        sys.stdout.write(
                            f"{datetime.now().isoformat()} {print_prefix}:{line}"
                        )
                        sys.stdout.flush()
            if proc.wait() != 0:
                nonlocal n_failed
                with file_lock:
                    n_failed += 1
                    fail_file.write(cmd + "\n")
                    fail_file.flush()
            q.task_done()

    if cpus is not None:
        threads = [Thread(target=worker, kwargs={"t_id": i}) for i in range(cpus)]
    else:
        _gpus = list(map(int, gpus.split(",")))
        threads = [Thread(target=worker, kwargs={"gpu": i}) for i in _gpus]
    print(f"Starting threads at {datetime.now().isoformat()}")
    start_time = datetime.now()
    for t in threads:
        t.start()
    for t in threads:
        t.join()
    total_time = datetime.now() - start_time
    print(f"Finished at {datetime.now().isoformat()}")
    print(f"Total time: {total_time} ({total_time.total_seconds()} s)")
    print(f"{n_commands - n_failed} commands succeeded, {n_failed} failed")


if __name__ == "__main__":
    main()



"""Dataset classes and related functions."""

import copy
import logging
import os
import warnings
from dataclasses import dataclass, field
from functools import partial, reduce
from itertools import chain
from pathlib import Path
from typing import (
    Any,
    Collection,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    Union,
    overload,
)

import numpy as np
import pandas as pd
from numpy.lib.arraysetops import setdiff1d
from omegaconf import MISSING, OmegaConf
from sklearn.base import TransformerMixin
from sklearn.preprocessing import StandardScaler
from typing_extensions import Literal

from ertk.config import ERTKConfig, get_arg_mapping
from ertk.dataset.annotation import read_annotations
from ertk.dataset.features import find_features_file, read_features
from ertk.dataset.utils import get_audio_paths
from ertk.transform import group_transform
from ertk.utils import PathOrStr, clip_arrays, frame_arrays, pad_arrays, transpose_time

__all__ = [
    "Dataset",
    "CombinedDataset",
    "DatasetConfig",
    "DataLoadConfig",
    "DataSelector",
    "SubsetInfo",
    "MapGroups",
    "RemoveGroups",
    "CorpusInfo",
    "load_multiple",
    "load_datasets_config",
]

logger = logging.getLogger(__name__)


@dataclass
class MapGroups:
    """Defined a mapping between categorical groups."""

    map: Dict[str, str] = MISSING


@dataclass
class RemoveGroups:
    """Defines a set of groups to keep or remove from a dataset."""

    drop: List[str] = field(default_factory=list)
    keep: List[str] = field(default_factory=list)


@dataclass
class SubsetInfo:
    """Defines a subset of a dataset."""

    clips: str = MISSING
    description: str = ""


@dataclass
class DataSelector(ERTKConfig):
    """Defines a selection of data to keep."""

    subset: str = ""
    groups: Dict[str, RemoveGroups] = field(default_factory=dict)


@dataclass
class CorpusInfo(ERTKConfig):
    """Defines information about a dataset."""

    name: str = MISSING
    description: str = ""
    annotations: List[str] = field(default_factory=list)
    partitions: List[str] = field(default_factory=list)
    ratings: List[str] = field(default_factory=list)
    subsets: Dict[str, SubsetInfo] = field(default_factory=dict)
    default_subset: str = "all"
    features_dir: str = "features"


@dataclass
class DatasetConfig(ERTKConfig):
    """Defines a dataset configuration."""

    path: str = MISSING
    features: Optional[str] = None
    read_kwargs: Dict[str, Any] = field(default_factory=dict)
    subset: str = "default"
    map_groups: Dict[str, MapGroups] = field(default_factory=dict)
    remove_groups: Dict[str, RemoveGroups] = field(default_factory=dict)
    rename_annotations: Dict[str, str] = field(default_factory=dict)
    select: Optional[DataSelector] = None
    clip_seq: Optional[int] = None
    pad_seq: Optional[int] = None


@dataclass
class DataLoadConfig(ERTKConfig):
    """Defines a configuration for loading one or more datasets."""

    datasets: Dict[str, DatasetConfig] = MISSING
    features: Optional[str] = None
    label: Optional[str] = None
    map_groups: Dict[str, MapGroups] = field(default_factory=dict)
    remove_groups: Dict[str, RemoveGroups] = field(default_factory=dict)
    select: Optional[DataSelector] = None
    clip_seq: Optional[int] = None
    pad_seq: Optional[int] = None


_AUDIO_PATH_KEY = "_audio_path"


class Dataset:
    """Class representing a generic dataset, consisting of a set of
    features and optional partitions and annotations. Has various
    preprocessing methods.

    An annotation is a scalar value for all instances in the dataset.

    A partition is a partition of instances into disjoint groups (e.g.
    speakers). A partition should be complete in that each instance is
    in exactly one group in the partition. Each partition has a
    corresponding annotation with the same name, using the group names
    as categorical annotations.

    Parameters
    ----------
    corpus_info: Pathlike or str, optional
        Path to corpus info in YAML format.
    features: Pathlike or str, optional
        Path to features file, or unique name of features in corpus
        features directory.
    subset: str, optional
        The subset of instances to use.
    """

    _annotations: pd.DataFrame
    _corpus: str = ""
    _feature_names: List[str]
    _names: pd.Index
    _partitions: Set[str]
    _subset: str = ""
    _subsets: Dict[str, List[str]]
    _x: np.ndarray
    _label_annot: str = ""
    _ratings: Dict[str, pd.DataFrame]
    _description: str = ""

    # "Private" vars
    _default_subset: str = ""
    _features: str = ""
    _features_path: Optional[Path] = None
    _features_dir: Path = Path()
    _subset_paths: Dict[str, Path]

    def __init__(
        self,
        corpus_info: PathOrStr,
        features: Optional[PathOrStr] = None,
        subset: str = "default",
        label: str = "label",
    ):
        self._init()
        self.init_corpus_info(corpus_info)
        self.use_subset(subset)
        if features is not None:
            self.update_features(features)
        self._label_annot = label

    def _init(self) -> None:
        self._annotations = pd.DataFrame()
        self._feature_names = []
        self._names = pd.Index([])
        self._partitions = set()
        self._subsets = {}
        self._subset_paths = {}
        self._x = np.empty((0, 0), dtype=np.float32)
        self._ratings = {}

    def _copy_from_dataset(self, other: "Dataset") -> None:
        self._annotations = other._annotations.copy()
        self._corpus = other._corpus
        self._default_subset = other._default_subset
        self._feature_names = other._feature_names.copy()
        self._features = other._features
        self._features_path = other._features_path
        self._features_dir = other._features_dir
        self._names = other._names.copy()
        self._partitions = other._partitions.copy()
        self._subset = other._subset
        self._subset_paths = other._subset_paths.copy()
        self._subsets = copy.deepcopy(other._subsets)
        self._x = other._x.copy()
        self._label_annot = other._label_annot
        self._ratings = copy.deepcopy(other._ratings)
        self._description = other._description
        if len(self._x.shape) == 1:
            # Non-contiguous array, so copy each contiguous sub-array
            for i in range(len(self._x)):
                self._x[i] = self._x[i].copy()

    def init_corpus_info(self, path: PathOrStr) -> None:
        """Initialise corpus metadata from YAML.

        Parameters
        ----------
        path: os.Pathlike or str
            The path to a YAML file containing corpus metadata.
        """
        logger.debug(f"Initialising corpus from {path}")

        path = Path(path)
        corpus_info = CorpusInfo.from_file(path)
        self._corpus = corpus_info.name
        self._description = corpus_info.description
        self._features_dir = path.parent / corpus_info.features_dir
        self._default_subset = corpus_info.default_subset
        for subset, subset_info in corpus_info.subsets.items():
            clips_file = path.parent / subset_info.clips
            if clips_file.suffix == "":
                clips_file = clips_file.with_suffix(".txt")
            self._subset_paths[subset] = clips_file
            name_to_path = {x.stem: str(x) for x in get_audio_paths(clips_file)}
            self.subsets[subset] = list(name_to_path.keys())
            if subset == "all":
                self._names = pd.Index(self.subsets["all"])
                # This will initialise the index of self.annotations
                self._annotations.index = pd.Index(self.subsets["all"])
                self.update_annotation(_AUDIO_PATH_KEY, list(name_to_path.values()))
        for partition in corpus_info.partitions:
            self.update_annotation(
                partition, path.parent / f"{partition}.csv", dtype="category"
            )
        for annotation in corpus_info.annotations:
            self.update_annotation(annotation, path.parent / f"{annotation}.csv")
        for ratings in corpus_info.ratings:
            self.update_ratings(ratings, path.parent / f"{ratings}.csv")
        self.update_annotation(
            "corpus", [self.corpus] * len(self._names), dtype="category"
        )

    def _verify_index(
        self, df_or_series: Union[pd.DataFrame, pd.Series], msg: Optional[str] = None
    ):
        msg = msg or "Incomplete index"
        try:
            df_or_series.loc[self.names]
        except KeyError as e:
            raise RuntimeError(msg) from e

    def _verify_annotations(self, msg: Optional[str] = "Incomplete annotation"):
        """Make sure there is a value for each instance for each annotation."""
        self._verify_index(self.annotations, msg=msg)

    def _verify_ratings(self):
        for name, df in self.ratings.items():
            try:
                self._verify_index(df, msg=f"Incomplete ratings for '{name}'")
            except RuntimeError as e:
                warnings.warn(str(e), RuntimeWarning)

    def _reset_categorical(self):
        """Make sure there are no extraneous categories in any
        partition.
        """
        for part in self.partitions:
            new = self._annotations.loc[self.names, part].cat.remove_unused_categories()
            self._annotations[part] = new

    def update_features(self, features: PathOrStr, **read_kwargs) -> None:
        """Update the features matrix and feature names for this dataset.

        Parameters
        ----------
        features: os.PathLike or str
            Path to a set of features or unique name of features in
            corpus features dir.
        **read_kwargs:
            Other arguments to pass to `read_features()`.
        """
        if features == "raw_audio":
            self._features = "raw_audio"
            features = self._subset_paths[self.subset or "all"]
        else:
            features = Path(features)
            if not features.exists():
                try:
                    features = find_features_file(
                        self._features_dir.glob(f"{features}.*")
                    )
                except FileNotFoundError as e:
                    raise FileNotFoundError(
                        f"Cannot find features '{features}' in directory "
                        f"'{self._features_dir}'"
                    ) from e
            self._features = features.stem
        self._features_path = features
        data = read_features(features, **read_kwargs)
        self._feature_names = data.feature_names

        assert len(self.names) > 0
        missing = self.names.difference(data.names)
        if len(missing) > 0:
            raise ValueError(
                f"Features at {features} don't contain all instances, these are"
                f"missing: {missing}"
            )
        # To avoid reordering the feature matrix, reorder self.names instead
        data_names = pd.Index(data.names)
        idx = data_names.isin(self.names)
        self._x = data.features[idx]
        self._names = data_names[idx]

    def use_subset(self, subset: str = "default") -> None:
        """Use a different subset of the instances.

        Parameters
        ----------
        subset: str
            Name of subset to use. Default is "default" which uses the
            default subset specified in corpus_info.
        """
        if subset == "default":
            subset = self._default_subset
        self._subset = subset
        self._names = pd.Index(self.subsets[subset])
        self._reset_categorical()
        self._verify_annotations()

    def get_idx_for_names(self, names: Collection[str]) -> np.ndarray:
        """Gets indices of instances corresponding to `names`.

        Parameters
        ----------
        names: collection of str
            The names to get indices for.

        Returns
        -------
        idx: np.ndarray
            The indices corresponding to `names`, in order.
        """
        return np.flatnonzero(self.names.isin(names))

    @overload
    def get_idx_for_split(
        self,
        split: Union[str, Dict[str, Collection[str]], DataSelector],
        return_complement: Literal[False] = False,
    ) -> np.ndarray:
        ...

    @overload
    def get_idx_for_split(
        self,
        split: Union[str, Dict[str, Collection[str]], DataSelector],
        return_complement: Literal[True],
    ) -> Tuple[np.ndarray, np.ndarray]:
        ...

    def get_idx_for_split(
        self,
        split: Union[str, Dict[str, Collection[str]], DataSelector],
        return_complement: bool = False,
    ) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
        """Gets indices of instances corresponding to the selection
        given by `split`.

        Parameters
        ----------
        split: str or dict
            Either a string containing the subset to select, groups to
            select, a path to such a config file, or a mapping object
            containing the groups to select.
        return_complement: bool
            If True, return a tuple of `idx`, `comp_idx` where
            `comp_idx` contains the complement indices not in the split.

        Returns
        -------
        idx: np.ndarray
            The corresponding indices.
        comp_idx: np.ndarray, optional
            If `return_complement` is True, this is also returned and
            contains the complement indices.
        """
        subset_idx = None
        if isinstance(split, str):
            if ":" not in split and not Path(split).exists():  # is a subset
                return self.get_idx_for_names(self.subsets[split])
            mapping = get_arg_mapping(split)
        elif isinstance(split, dict):
            mapping = split
        else:
            if OmegaConf.get_type(split) is None:
                raise TypeError("Config is of incorrect type")
            if split.subset:
                subset_idx = self.get_idx_for_names(self.subsets[split.subset])
            mapping = {
                k: v.keep if v.keep else set(self.get_group_names(k)) - set(v.drop)
                for k, v in split.groups.items()
            }

        indices_parts: Dict[str, np.ndarray] = {}
        for part_name in mapping:
            sel = mapping[part_name]
            sel = [sel] if isinstance(sel, str) else sel
            group_names = self.get_group_names(part_name)
            sel_idx = [group_names.index(x) for x in sel]
            indices_parts[part_name] = np.flatnonzero(
                np.isin(self.get_group_indices(part_name), sel_idx)
            )

        all_idx = np.arange(len(self))
        indices = reduce(
            partial(np.intersect1d, assume_unique=True), indices_parts.values(), all_idx
        )
        if subset_idx is not None:
            indices = np.intersect1d(indices, subset_idx, assume_unique=True)
        comp_indices = setdiff1d(all_idx, indices, assume_unique=True)
        if return_complement:
            return np.sort(indices), np.sort(comp_indices)
        else:
            return np.sort(indices)

    def update_annotation(
        self,
        annot_name: str,
        annotations: Union[PathOrStr, Mapping[str, Any], Sequence[Any], pd.Series],
        dtype: Optional[Union[Type, Literal["category"]]] = None,
    ) -> None:
        """Add or update an annotation.

        Parameters
        ----------
        annot_name: str
            The name of the annotation.
        annotations: PathLike, str, mapping, DataFrame, Series or sequence
            Annotations to add, similar to update_partition(). If
            PathLike or str, annotations are read from a CSV. If a dict,
            should be of the form {instance: annotation}. If a list,
            should have an annotation for each instance.
        dtype: type, optional
            The type of annotations for reading from CSV file. If the
            literal "category" is given the annotations are converted to
            the Pandas categorical dtype.
        """
        if isinstance(annotations, (os.PathLike, str)):
            series = read_annotations(annotations, dtype=dtype)
        elif isinstance(annotations, Mapping):
            series = pd.DataFrame.from_dict(annotations, orient="index")
            series = series.squeeze("columns")
        elif isinstance(annotations, pd.Series):
            series = annotations
        else:
            series = pd.Series(annotations, index=self.names)

        logger.debug(f"Updating annotation {annot_name} ({len(series)} values).")

        try:
            self._verify_index(series)
        except RuntimeError as e:
            raise RuntimeError(f"Annotation {annot_name} is incomplete") from e
        self._annotations[annot_name] = series
        if (
            dtype == "category"
            or (dtype is not None and issubclass(dtype, str))
            or series.dtype == object
            or series.dtype == pd.CategoricalDtype
        ):
            self.partitions.add(annot_name)
            self._annotations[annot_name] = self._annotations[annot_name].astype(
                "category"
            )

    def update_ratings(
        self,
        rating_set: str,
        ratings: Union[PathOrStr, Mapping[str, Mapping[str, Any]], pd.Series],
    ) -> None:
        """Update a set of ratings.

        Parameters
        ----------
        rating_set: str
            The name for this set of ratings.
        ratings: PathLike, str, mapping, DataFrame, Series
            The ratings to add. Must have a joint index where
        """
        if isinstance(ratings, (os.PathLike, str)):
            df = pd.read_csv(ratings, converters={0: str, 1: str})
            df = df.set_index(list(df.columns[0:2]))
        elif isinstance(ratings, pd.DataFrame):
            df = ratings
        elif isinstance(ratings, pd.Series):
            df = ratings.to_frame()

        self.ratings[rating_set] = df
        self._verify_ratings()

    def remove_ratings(self, rating_set: str) -> None:
        """Delete a set of ratings for this dataset.

        Parameters
        ----------
        rating_set: str
            The name of a set of ratings.
        """
        del self.ratings[rating_set]

    def map_groups(self, part_name: str, mapping: Mapping[str, str]) -> None:
        """Map group names in a partition.

        Parameters
        ----------
        part_name: str
            Name of partition.
        mapping: dict
            Group name mapping.
        """
        logger.debug(f"Mapping {part_name}: {mapping}")

        annot = self._annotations[part_name]
        try:
            new = annot.cat.rename_categories(mapping)
        except ValueError:
            # Not a 1-1 mapping
            _map = {x: x for x in annot.cat.categories}
            _map.update(mapping)
            new = annot.map(_map).astype("category")
        new = new.cat.reorder_categories(new.cat.categories.sort_values())
        self._annotations[part_name] = new

        # groups = self.annotations[part_name]
        # self.update_annotation(
        #     part_name, groups.map(lambda x: mapping.get(x, x)), dtype=str
        # )

    def remove_groups(
        self,
        part_name: str,
        *,
        drop: Collection[str] = None,
        keep: Collection[str] = None,
    ) -> None:
        """Remove instances corresponding to groups from the given
        partition.

        Parameters
        ----------
        part_name: str
            The partition name.
        drop: collection of str
            The groups to remove in given partition.
        keep: collection of str
            The groups to keep in given partition.
        """
        drop = set([] if drop is None else drop)
        keep = set([] if keep is None else keep)

        if (len(drop) == 0) == (len(keep) == 0):
            raise ValueError("Exactly one of drop and keep should be non-empty.")

        all_groups = set(self.get_group_names(part_name))
        unknown = drop.union(keep) - all_groups
        if len(unknown) > 0:
            raise ValueError(f"Cannot drop or keep {unknown} in partition {part_name}.")

        if len(keep) == 0:
            keep = all_groups - drop

        if len(drop) == 0:
            drop = all_groups - keep

        logger.debug(
            f"Dropping {len(drop)} and keeping {len(keep)} groups from "
            f"partition {part_name}."
        )

        new = self._annotations[part_name].cat.remove_categories(list(drop))
        self._annotations[part_name] = new

        keep_names = new.index[new.isin(keep)]
        self.remove_instances(keep=keep_names.intersection(self.names))

    def map_and_select(
        self,
        map: Mapping[str, Mapping[str, str]],
        select: Mapping[str, Union[str, Collection[str]]],
        remove: Mapping[str, Union[str, Collection[str]]],
    ) -> None:
        """Convenience function for mapping one or more partitions and
        then selecting one or more groups.

        Parameters
        ----------
        map: mapping
            The groups mapping. May have one or more partitions.
        select: mapping
            Mapping from partitions to groups to select.
        """
        for part in map:
            if part not in self.partitions:
                warnings.warn(
                    f"Partition {part} cannot be mapped as it is not in the dataset."
                )
                continue
            self.map_groups(part, map[part])
        for part in set(select).union(remove):
            if part not in self.partitions:
                warnings.warn(
                    f"Partition {part} cannot be selected/dropped as it is not in the "
                    "dataset."
                )
                continue
            keep = select.get(part, [])
            drop = remove.get(part, [])
            if isinstance(keep, str):
                keep = [keep]
            if isinstance(drop, str):
                drop = [drop]
            self.remove_groups(part, keep=keep, drop=drop)

    def rename_annotation(self, old_name: str, new_name: str) -> None:
        """Renames an annotation. This is useful for example if you want
        to select a different labelling to use. If an annotation already
        exists with the given new_name, then it is replaced
        destructively.

        Parameters
        ----------
        old_name: str
            The name of the annotation to rename.
        new_name: str
            The new name of the annotation.
        """
        self._annotations.rename(columns={old_name: new_name}, inplace=True)
        if old_name in self.partitions:
            self.partitions.add(new_name)
            self.partitions.remove(old_name)

    def remove_annotation(self, annot_name: str) -> None:
        """Removes a set of annotations from the dataset. This does not
        remove any instances.

        Parameters
        ----------
        annot_name: str
            Annotation name.
        """
        del self._annotations[annot_name]
        # Use difference_update in case annot_name is not already a partition
        self.partitions.difference_update({annot_name})

    def get_annotations(self, annot_name: str) -> np.ndarray:
        """Get a list of annotations, one for each instance currently in
        the dataset.

        Parameters
        ----------
        annot_name: str
            Annotation name.

        Returns
        -------
        A pd.Series of values, one for each instance in the datset, in
        the same order they appear in names and x.
        """
        return self.annotations.loc[self.names, annot_name].to_numpy()

    def get_audio_paths(self) -> np.ndarray:
        return self.get_annotations(_AUDIO_PATH_KEY)

    def get_ratings(self, name: str, rating_set: str = "ratings") -> pd.Series:
        """Get per-annotator ratings of a specified column, for this
        dataset.

        Parameters
        ----------
        name: str
            The name of the rating column to get.
        rating_set: str
            The name of the rating set to get.

        Returns
        -------
        pd.Series
            Pandas Series with (name, rater) multiindex.
        """
        return self.ratings[rating_set].loc[self.names, name]

    def annotation_type(self, annot_name: str) -> Type:
        return self.annotations[annot_name].dtype

    def get_group_indices(self, annot_name: str) -> np.ndarray:
        """Gets the group indices (i.e. indices into the groups array)
        for a given partition.

        Parameters
        ----------
        annot_name: str
            The partition name.

        Returns
        -------
        A NumPy array of group indices for each instance in the dataset.
        """
        # _, idx = np.unique(self.get_annotations(annot_name), return_inverse=True)
        idx = self.annotations.loc[self.names, annot_name].cat.codes.to_numpy(np.int64)
        return idx

    def get_group_counts(self, annot_name: str) -> np.ndarray:
        """Get group counts for a partition.

        Parameters
        ----------
        annot_name: str
            The partition name.

        Returns
        -------
        A NumPy array of counts for the corresponding group in this
        partition.
        """
        counts = (
            self.annotations.loc[self.names, annot_name]
            .value_counts(sort=False)
            .to_numpy()
        )
        return counts

    def get_group_names(self, annot_name: str) -> List[str]:
        """Get the names of groups in a partition.

        Parameters
        ----------
        annot_name: str
            Annotation name.
        """
        return list(self.annotations.loc[self.names, annot_name].cat.categories)

    def remove_instances(
        self, *, drop: Collection[str] = None, keep: Collection[str] = None
    ):
        """Remove instances from dataset. Recalculate annotations,
        partitions, etc.

        Parameters
        ----------
        drop: collection of str
            Instances to drop.
        keep: collection of str
            Instances to keep. Exactly one of drop and keep should be
            given.
        """
        drop = set([] if drop is None else drop)
        keep = set([] if keep is None else keep)

        if (len(drop) == 0) == (len(keep) == 0):
            raise ValueError("Exactly one of drop and keep should be given.")

        if len(keep) == 0:
            keep = self.names.difference(drop, sort=False)

        # Need idx for self.x, instead of ordered_intersect()
        idx = self.get_idx_for_names(keep)
        self._names = self.names[idx]
        if len(self._x) > 0:  # To avoid NumPy DeprecationWarning
            self._x = self.x[idx]
        self._reset_categorical()
        self._verify_annotations(msg="Incomplete annotations after removing instances.")

    def clone(self):
        return self.copy()

    def copy(self):
        new_dataset = type(self).__new__(type(self))
        new_dataset._copy_from_dataset(self)
        return new_dataset

    def normalise(
        self, partition: str, normaliser: TransformerMixin = StandardScaler()
    ):
        """Transforms the data matrix of this dataset in-place using
        some (offline) normalisation method.

        Parameters
        ----------
        normaliser:
            The transform to apply. Must implement fit_transform().
        partition: str
            The partition to apply in a per-group fashion. If "all",
            then perform global normalisation on all the data.
        """

        if partition == "all":
            groups = np.zeros(len(self.x), dtype=int)
        elif partition == "instance":
            groups = np.arange(len(self.x))
        else:
            groups = self.get_group_indices(partition)
        group_transform(self.x, groups, normaliser, inplace=True)

    def pad_arrays(self, pad: int = 32):
        """Pads each array to the nearest multiple of `pad` greater than
        the array size. Assumes axis 0 of x is time.
        """
        logger.info(f"Padding array lengths to nearest multiple of {pad}.")
        self._x = pad_arrays(self.x, pad=pad)

    def clip_arrays(self, length: int):
        """Clips each array to the specified maximum length."""
        logger.info(f"Clipping arrays to max length {length}.")
        self._x = clip_arrays(self.x, length=length, copy=False)

    def frame_arrays(
        self,
        frame_size: int,
        frame_shift: int,
        max_frames: Optional[int] = None,
    ):
        """Create a sequence of frames from the raw signal."""
        logger.info(f"Framing arrays with size {frame_size} and shift {frame_shift}.")
        self._x = frame_arrays(
            self._x,
            frame_size=frame_size,
            frame_shift=frame_shift,
            max_frames=max_frames,
        )

    def transpose_time(self):
        """Transpose the time and feature axis of each instance."""
        logger.info("Transposing time and feature axis of data.")
        self._x = transpose_time(self._x)

    @property
    def corpus(self) -> str:
        """The corpus this dataset represents."""
        return self._corpus

    @property
    def description(self) -> str:
        """The descriptive name of this corpus"""
        return self._description

    @property
    def n_instances(self) -> int:
        """Number of instances in this dataset."""
        return len(self.names)

    @property
    def feature_names(self) -> List[str]:
        """List of feature names."""
        return self._feature_names

    @property
    def n_features(self) -> int:
        """Number of features."""
        return len(self.feature_names)

    @property
    def speaker_names(self) -> List[str]:
        """List of unique speakers in this dataset."""
        return self.get_group_names("speaker")

    @property
    def speaker_counts(self) -> np.ndarray:
        """Number of instances for each speaker."""
        return self.get_group_counts("speaker")

    @property
    def speaker_indices(self) -> np.ndarray:
        """Indices into speakers array of corresponding speaker for each
        instance.
        """
        return self.get_group_indices("speaker")

    @property
    def speakers(self) -> np.ndarray:
        return self.get_annotations("speaker")

    @property
    def n_speakers(self) -> int:
        return len(self.speaker_names)

    @property
    def partitions(self) -> Set[str]:
        """Partitions in this dataset."""
        return self._partitions

    @property
    def annotations(self) -> pd.DataFrame:
        """Full annotation matrix for dataset."""
        return self._annotations.loc[self.names]

    @property
    def ratings(self) -> Dict[str, pd.DataFrame]:
        """Full ratings for dataset."""
        return self._ratings

    @property
    def names(self) -> pd.Index:
        """List of instance names."""
        return self._names

    @property
    def subset(self) -> str:
        """Name of clip subset used."""
        return self._subset

    @property
    def subsets(self) -> Dict[str, List[str]]:
        """Dict from subset name to set of clip names."""
        return self._subsets

    @property
    def x(self) -> np.ndarray:
        """The data matrix."""
        return self._x

    def __len__(self) -> int:
        return self.n_instances

    def update_labels(
        self, labels: Union[PathOrStr, Mapping[str, str], Sequence[str], pd.Series]
    ):
        self.update_annotation(self.label_annot, labels, dtype=str)

    def map_classes(self, mapping: Mapping[str, str]):
        """Modifies classses based on the mapping in map. Keys not
        corresponding to classes are ignored. The new classes will be
        sorted lexicographically.
        """
        self.map_groups(self.label_annot, mapping)

    def remove_classes(
        self, *, drop: Collection[str] = None, keep: Collection[str] = None
    ):
        """Remove instances with labels not in `keep`."""
        self.remove_groups(self.label_annot, drop=drop, keep=keep)

    @property
    def classes(self) -> List[str]:
        """List of unique class labels."""
        return self.get_group_names(self.label_annot)

    @property
    def n_classes(self) -> int:
        """Number of unique classes."""
        return len(self.classes)

    @property
    def class_counts(self) -> np.ndarray:
        """Number of instances for each class."""
        return self.get_group_counts(self.label_annot)

    @property
    def labels(self) -> np.ndarray:
        """List of labels for instances."""
        return self.get_annotations(self.label_annot)

    @property
    def label_annot(self) -> str:
        """The annotation used as target label."""
        return self._label_annot

    @label_annot.setter
    def label_annot(self, val: str) -> None:
        self._label_annot = val

    @property
    def y(self) -> np.ndarray:
        """The class label array; one label per instance."""
        if self.label_annot in self.partitions:
            return self.get_group_indices(self.label_annot)
        else:
            return self.get_annotations(self.label_annot)

    def __str__(self):
        s = f"Corpus: {self.corpus}\n"
        s += f"Description: {self.description}\n"
        for part in sorted(self.partitions):
            if part == _AUDIO_PATH_KEY:
                continue
            group_names = self.get_group_names(part)
            s += f"partition '{part}' ({len(group_names)} groups)"
            n_invalid = self.annotations.loc[self.names, part].isna().sum()
            if n_invalid > 0:
                n_valid = len(self) - n_invalid
                s += f" (incomplete [{n_valid} valid])"
            s += "\n"
            if len(group_names) <= 20:
                s += f"\t{dict(zip(group_names, self.get_group_counts(part)))}\n"
        for annot in sorted(self.annotations):
            if annot in self.partitions:
                continue
            s += f"annotation '{annot}'\n"
            annotations = self.get_annotations(annot)
            s += f"\tmin: {annotations.min():.3f}, max: {annotations.max():.3f}, "
            s += f"mean: {annotations.mean():.3f}\n"
        s += f"{self.n_instances} instances\n"
        s += "Subsets:\n" if self.subsets else ""
        for subset, names in self.subsets.items():
            s += "\t*" if subset == self.subset else "\t "
            s += f"{subset}: {len(names)} instances\n"
        s += f"using {self._features} ({self.n_features} features)\n"
        if self.x.dtype == object or len(self.x.shape) == 3:
            lengths = [len(x) for x in self.x]
            s += "Sequences:\n"
            s += f"min length: {np.min(lengths)}\n"
            s += f"mean length: {np.mean(lengths)}\n"
            s += f"max length: {np.max(lengths)}\n"
        return s


class CombinedDataset(Dataset):
    """A dataset that joins one or more individual datasets together and
    merges annotations.
    """

    def __init__(self, *datasets: Dataset):
        if len(datasets) == 0:
            raise ValueError("No datasets provided.")
        if len(datasets) == 1:
            self._copy_from_dataset(datasets[0])
            return

        self._init()

        combined_str = ", ".join(d.corpus for d in datasets)
        logger.info(f"Combining {combined_str}")
        self._corpus = "combined"
        self._description = f"Combined datasets ({combined_str})"
        self._names = pd.Index([f"{d.corpus}_{n}" for d in datasets for n in d.names])
        self._feature_names = datasets[0].feature_names
        self._features = datasets[0]._features
        self._x = np.concatenate([d.x for d in datasets])
        self.update_annotation(
            "corpus", [d.corpus for d in datasets for _ in d.names], dtype="category"
        )

        all_speakers = {}
        for d in datasets:
            if "speaker" not in d.annotations.columns:
                all_speakers.update(
                    {f"{d.corpus}_{k}": f"{d.corpus}_unknown" for k in d.names}
                )
            else:
                speakers = d.annotations["speaker"]
                all_speakers.update(
                    {f"{d.corpus}_{k}": f"{d.corpus}_{v}" for k, v in speakers.items()}
                )
        self.update_annotation("speaker", all_speakers)

        # TODO: handle case when only some datasets have a given annotation?
        all_annotations = set(chain(*(d.annotations.columns for d in datasets)))
        all_annotations -= {"speaker", "corpus"}  # assumed to be per-dataset
        common_annotations = {
            x
            for x in all_annotations
            if all(x in d.annotations.columns for d in datasets)
        }
        logger.info(f"All annotations: {all_annotations}")
        logger.info(f"Common annotations: {common_annotations}")
        for annot_name in common_annotations:
            combined_annot = {}
            for dataset in datasets:
                annotations = dataset.annotations[annot_name]
                combined_annot.update(
                    {f"{dataset.corpus}_{k}": v for k, v in annotations.items()}
                )
            self.update_annotation(annot_name, combined_annot)

        # Non-common categorical annotations
        for annot_name in all_annotations - common_annotations:
            if not any(annot_name in d.partitions for d in datasets):
                continue
            combined_annot = {}
            for d in datasets:
                if annot_name not in d.annotations:
                    combined_annot.update({f"{d.corpus}_{k}": None for k in d.names})
                else:
                    annotations = d.annotations[annot_name]
                    combined_annot.update(
                        {f"{d.corpus}_{k}": v for k, v in annotations.items()}
                    )
            self.update_annotation(annot_name, combined_annot)

        if any(d.label_annot == "" for d in datasets):
            return

        if any(d.label_annot != datasets[0].label_annot for d in datasets):
            # Try and unify label annotations with different names
            warnings.warn(
                "label annotation differs between some datasets, attempting to unify."
            )
            combined_labels = {}
            for d in datasets:
                for name in d.names:
                    label = d.annotations[d.label_annot][name]
                    combined_labels[f"{d.corpus}_{name}"] = label
            self.update_annotation("_combined_labels", combined_labels)
            self.label_annot = "_combined_labels"
        else:
            self.label_annot = next(d.label_annot for d in datasets)

    @property
    def corpus_names(self) -> List[str]:
        """List of corpora in this CombinedDataset."""
        return self.get_group_names("corpus")

    @property
    def corpus_indices(self) -> np.ndarray:
        """Indices into corpora list of corresponding corpus for each
        instance.
        """
        return self.get_group_indices("corpus")


def load_multiple(
    corpus_files: Iterable[PathOrStr],
    features: Optional[str] = None,
    subsets: Union[str, Mapping[str, str]] = "default",
    label: Union[str, Mapping[str, str]] = "label",
    **read_kwargs,
) -> CombinedDataset:
    """Load one or more datasets with the given features.

    Parameters
    ----------
    corpus_files: iterable
        The corpus description YAML files to load.
    features: str
        A common set of features to load. This will be found in the
        features directory corresponding to each corpus.
    subsets: str or dict
        A subset name common to all datasets (e.g. "all", "default") or
        a mapping from dataset name to subset name.
    **read_kwargs:
        Other args to pass to feature loading.
    """
    corpus_files = list(corpus_files)
    if len(corpus_files) == 0:
        raise RuntimeError("No corpus metadata files given.")
    datasets = []
    for file in corpus_files:
        logger.info(f"Loading {file}")
        dataset = Dataset(file, subset="all")
        if isinstance(label, str):
            setattr(dataset, "label_annot", label)
        else:
            setattr(dataset, "label_annot", label.get(dataset.corpus, "label"))
        if isinstance(subsets, str):
            dataset.use_subset(subsets)
        else:
            dataset.use_subset(subsets.get(dataset.corpus, "default"))
        if features is not None:
            dataset.update_features(features, **read_kwargs)
        datasets.append(dataset)
    return CombinedDataset(*datasets)


def load_datasets_config(config: Union[PathOrStr, DataLoadConfig]) -> Dataset:
    """Load one or more datasets from a DataLoadConfig.

    Parameters
    ----------
    config: DataLoadConfig
        The data loading config.

    Returns
    -------
    Dataset
        A dataset that combines one or more datasets according to the
        given config.
    """
    if isinstance(config, (str, os.PathLike)):
        config = DataLoadConfig.from_file(config)
    datasets: List[Dataset] = []
    for corpus, dataset_conf in config.datasets.items():
        logger.info(f"Loading {dataset_conf.path}")
        dataset = Dataset(dataset_conf.path, subset=dataset_conf.subset)
        dataset._corpus = corpus
        for part, mapping in dataset_conf.map_groups.items():
            dataset.map_groups(part, mapping.map)
        for part, remove in dataset_conf.remove_groups.items():
            dataset.remove_groups(part, drop=remove.drop, keep=remove.keep)
        for a1, a2 in dataset_conf.rename_annotations.items():
            # TODO: check for collisions
            dataset.rename_annotation(a1, a2)
        if dataset_conf.features:
            dataset.update_features(dataset_conf.features, **dataset_conf.read_kwargs)
        elif config.features:
            dataset.update_features(config.features, **dataset_conf.read_kwargs)
        if dataset_conf.clip_seq:
            dataset.clip_arrays(dataset_conf.clip_seq)
        if dataset_conf.pad_seq:
            dataset.pad_arrays(dataset_conf.pad_seq)
        datasets.append(dataset)
    if len(config.datasets) > 1:
        dataset = CombinedDataset(*datasets)
        del datasets
    else:
        dataset = datasets[0]
    for part, mapping in config.map_groups.items():
        dataset.map_groups(part, mapping.map)
    for part, remove in config.remove_groups.items():
        dataset.remove_groups(part, drop=remove.drop, keep=remove.keep)
    if config.clip_seq:
        dataset.clip_arrays(config.clip_seq)
    if config.pad_seq:
        dataset.pad_arrays(config.pad_seq)
    if config.label:
        dataset.label_annot = config.label
    return dataset



import pytest

from ertk.cli.cli.train import main


def test_noargs():
    with pytest.raises(SystemExit) as excinfo:
        main([])
    assert excinfo.value.code == 2


def test_badargs():
    with pytest.raises(SystemExit) as excinfo:
        main(["x", "y"])
    assert excinfo.value.code == 2



"""Multi-task learning models."""

import warnings

import liblinear.liblinearutil as liblinearutil
import numpy as np
import scipy.sparse as sp
from sklearn.base import BaseEstimator, ClassifierMixin, clone
from sklearn.exceptions import ConvergenceWarning
from sklearn.linear_model import LogisticRegression
from sklearn.model_selection import cross_val_predict
from sklearn.multioutput import ClassifierChain
from sklearn.preprocessing import OneHotEncoder
from sklearn.utils import check_random_state
from sklearn.utils.validation import check_is_fitted

__all__ = ["BaseMTFL", "MTFLClassifier", "MultiDimensionalCC"]

_EPS = 1e-15


class BaseMTFL(BaseEstimator):
    """Linear regressor based on MTFL. Code adapted from
    https://github.com/argyriou/multi_task_learning

    Parameters
    ----------
    gamma: float
        Gamma value.
    eps_init: float
        Initial value for epsilon.
    n_iters: int
        Number of iterations per step.
    """

    def __init__(
        self,
        gamma: float = 1,
        epsilon: float = 1e-4,
        n_iters: int = 10,
        loss: str = "logistic",
    ):
        self.gamma = gamma
        self.epsilon = epsilon
        self.n_iters = n_iters
        self.loss = loss

    def _loss(self, X, y, task_idxs):
        num_tasks = len(task_idxs)
        W = np.zeros((X.shape[1], num_tasks))
        cost = 0
        err = 0
        reg = 0
        for t_id in range(num_tasks):
            task_x = X[task_idxs[t_id]]
            task_y = y[task_idxs[t_id]]

            if self.loss == "logistic":
                model = liblinearutil.train(
                    task_y, task_x, f"-s 0 -e 1e-3 -c {self.gamma} -q"
                )
                w, b = model.get_decfun()
            elif self.loss == "svm":
                model = liblinearutil.train(
                    task_y, task_x, f"-s 2 -e 1e-3 -c {self.gamma} -q"
                )
                w, b = model.get_decfun()
            elif self.loss == "ols":
                K = task_x.T @ task_x
                mat_inv = np.linalg.inv(K + self.gamma * np.eye(K.shape[0]))
                w = mat_inv @ task_x.T @ task_y

                # K = task_x @ task_x.T  # Linear kernel
                # mat_inv = np.linalg.inv(K + self.gamma * np.eye(K.shape[0]))
                # a = mat_inv @ task_x.T @ task_y
                # cost += self.gamma * task_y @ a
                # err += self.gamma**2 * a.T @ a
                # W[:, t_id] = task_x.T @ a
            w = np.asarray(w)
            W[:, t_id] = w
            _cost = self.gamma * w.T @ w
            cost += _cost
        reg = cost - err
        return W, cost, err, reg

    def Dmin(self, a):
        return a / a.sum()

    def Dmin_eps(self, a, eps):
        return self.Dmin(np.sqrt(a**2 + eps))

    def fit(self, X, y, tasks, sample_weight=None):
        self._fit_feat(X, y, tasks)
        return self

    def _fit_feat(self, X, y, tasks):
        task_ids = np.unique(tasks)
        task_idxs = [np.flatnonzero(tasks == t) for t in task_ids]
        dim = X.shape[1]
        Dini = np.eye(dim) / dim

        def fmeth(a):
            _a = np.zeros_like(a)
            _a[a > _EPS] = 1 / a[a > _EPS]
            return _a

        # train_alternating
        U, s, _ = np.linalg.svd(Dini, hermitian=True)
        fS = np.sqrt(fmeth(s))
        fS[fS > _EPS] = 1 / fS[fS > _EPS]
        fD_isqrt = U @ np.diag(fS) @ U.T
        _costs = np.empty((self.n_iters, 3))
        for i in range(self.n_iters):
            W, cost, err, reg = self._loss(X @ fD_isqrt, y, task_idxs)
            W = fD_isqrt @ W

            _costs[i] = [cost, err, reg]

            U, s, _ = np.linalg.svd(W)
            if dim > len(task_ids):
                s = np.r_[s, np.zeros(dim - len(task_ids))]
            Smin = self.Dmin_eps(s, eps=self.epsilon)
            D = U @ np.diag(Smin) @ U.T

            U, s, _ = np.linalg.svd(D, hermitian=True)
            fS = np.sqrt(fmeth(s))
            fS[fS > _EPS] = 1 / fS[fS > _EPS]
            fD_isqrt = U @ np.diag(fS) @ U.T
        # train_alternating

        s = np.linalg.svd(D, compute_uv=False)
        _costs[:, [0, 2]] = (
            _costs[:, [0, 2]] + self.gamma * self.epsilon * fmeth(s).sum()
        )
        self.mineps_ = self.epsilon
        self.W_ = W
        self.D_ = D
        self.costs_ = [_costs]

    def predict(self, X):
        return X @ self.W_


class MTFLClassifier(BaseMTFL, ClassifierMixin):
    """Classifier based on MTFL. Code adapted from
    https://github.com/argyriou/multi_task_learning

    Parameters
    ----------
    gamma: float
        Gamma value.
    eps_init: float
        Initial value for epsilon.
    n_iters: int
        Number of iterations per step.
    loss: str
        Loss to use.
    aggregate: str
        Method of aggregation. One of "logistic", "weighted",
        "unweighted".
    """

    def __init__(
        self,
        gamma: float = 1,
        epsilon: float = 1e-4,
        n_iters: int = 10,
        loss: str = "svm",
        aggregate: str = "weighted",
    ):
        self.gamma = gamma
        self.epsilon = epsilon
        self.n_iters = n_iters
        self.loss = loss
        self.aggregate = aggregate

    def fit(self, X, y, tasks, sample_weight=None):
        if len(np.unique(y)) > 2:
            raise ValueError("`y` should have only two classes")
        y = y.copy()
        y[y == 0] = -1
        self._fit_feat(X, y, tasks)
        if self.aggregate == "logistic":
            self.logistic_ = LogisticRegression(penalty="none")
            warnings.simplefilter("ignore", ConvergenceWarning)
            self.logistic_.fit(super().predict(X), y)
            warnings.simplefilter("default", ConvergenceWarning)
        return self

    def decision_function(self, X):
        preds = super().predict(X)
        if self.aggregate == "logistic":
            return self.logistic_.decision_function(preds)
        return preds.mean(-1)

    def predict(self, X):
        preds = super().predict(X)
        if self.aggregate == "logistic":
            preds = self.logistic_.predict(preds)
        elif self.aggregate == "weighted":
            preds = np.sign(preds.mean(-1))
        elif self.aggregate == "unweighted":
            preds = np.sign(np.sign(preds).mean(-1))
        preds[preds == -1] = 0
        return preds


class MultiDimensionalCC(ClassifierChain):
    """Classifier chain for multi-dimensional output. Adapted from
    sklearn's ClassifierChain and _BaseChain.

    Parameters
    ----------
    base_estimator: estimator
        The base estimator from which the classifier chain is built.
    order: list of lists
        The order of the outputs in the classifier chain.
    """

    def __init__(
        self,
        base_estimator,
        *,
        order=None,
        cv=None,
        random_state=None,
        verbose: bool = False,
    ) -> None:
        super().__init__(base_estimator, order=order, cv=cv, random_state=random_state)
        self.verbose = verbose

    def fit(self, X, Y, **fit_params):
        X, Y = self._validate_data(X, Y, multi_output=True, accept_sparse=True)

        random_state = check_random_state(self.random_state)
        self.order_ = self.order
        if isinstance(self.order_, tuple):
            self.order_ = np.array(self.order_)

        if self.order_ is None:
            self.order_ = np.array(range(Y.shape[1]))
        elif isinstance(self.order_, str):
            if self.order_ == "random":
                self.order_ = random_state.permutation(Y.shape[1])
        elif sorted(self.order_) != list(range(Y.shape[1])):
            raise ValueError("invalid order")

        self.estimators_ = [clone(self.base_estimator) for _ in range(Y.shape[1])]

        encoder = OneHotEncoder(sparse=False)
        Y_onehot = encoder.fit_transform(Y[:, self.order_])
        cat_sizes = np.array([len(x) for x in encoder.categories_])
        cat_sizes_cumsum = np.concatenate([[0], cat_sizes.cumsum()])
        total_size = cat_sizes_cumsum[-1]

        if self.cv is None:
            Y_pred_chain = Y_onehot
            if sp.issparse(X):
                X_aug = sp.hstack((X, Y_pred_chain), format="lil")
                X_aug = X_aug.tocsr()
            else:
                X_aug = np.hstack((X, Y_pred_chain))
        elif sp.issparse(X):
            Y_pred_chain = sp.lil_matrix((X.shape[0], total_size))
            X_aug = sp.hstack((X, Y_pred_chain), format="lil")
        else:
            Y_pred_chain = np.zeros((X.shape[0], total_size))
            X_aug = np.hstack((X, Y_pred_chain))

        del Y_pred_chain

        for chain_idx, estimator in enumerate(self.estimators_):
            y = Y[:, self.order_[chain_idx]]
            estimator.fit(
                X_aug[:, : X.shape[1] + cat_sizes_cumsum[chain_idx]], y, **fit_params
            )
            if self.cv is not None and chain_idx < len(self.estimators_) - 1:
                col_start = X.shape[1] + cat_sizes_cumsum[chain_idx]
                col_end = X.shape[1] + cat_sizes_cumsum[chain_idx + 1]
                cv_result = cross_val_predict(
                    self.base_estimator, X_aug[:, :col_start], y=y, cv=self.cv
                )
                y_oh = OneHotEncoder(sparse=False).fit_transform(
                    cv_result.reshape(-1, 1)
                )
                if sp.issparse(X_aug):
                    X_aug[:, col_start:col_end] = y_oh
                else:
                    X_aug[:, col_start:col_end] = y_oh

        self.classes_ = [estimator.classes_ for estimator in self.estimators_]
        return self

    def predict(self, X):
        check_is_fitted(self)
        X = self._validate_data(X, accept_sparse=True, reset=False)
        Y_pred_chain = np.zeros((X.shape[0], len(self.estimators_)))

        X_aug = X
        for chain_idx, estimator in enumerate(self.estimators_):
            if chain_idx != 0:
                y_oh = OneHotEncoder(
                    sparse=False, categories=[self.classes_[chain_idx - 1]]
                ).fit_transform(Y_pred_chain[:, chain_idx - 1].reshape(-1, 1))
                if sp.issparse(X):
                    X_aug = sp.hstack((X_aug, y_oh))
                else:
                    X_aug = np.hstack((X_aug, y_oh))
            Y_pred_chain[:, chain_idx] = estimator.predict(X_aug)

        inv_order = np.empty_like(self.order_)
        inv_order[self.order_] = np.arange(len(self.order_))
        Y_pred = Y_pred_chain[:, inv_order]

        return Y_pred



"""Process the raw VIVAE dataset.

This assumes the file structure from the original data:
/.../
    full_set/
        *.wav
    core_set/
        *.wav
"""

from pathlib import Path

import click

from ertk.dataset import resample_audio, write_annotations, write_filelist


@click.command()
@click.argument(
    "input_dir", type=click.Path(exists=True, file_okay=False, path_type=Path)
)
@click.option("--resample/--noresample", default=True)
def main(input_dir: Path, resample: bool):
    """Process the VIVAE dataset at location INPUT_DIR."""
    paths = list(input_dir.glob("full_set/*.wav"))
    resample_dir = Path("resampled")
    if resample:
        resample_audio(paths, resample_dir)
    write_filelist(resample_dir.glob("*.wav"), "files_all")
    core_set = [
        resample_dir / x.name
        for x in paths
        if (input_dir / "core_set" / x.name).exists()
    ]
    write_filelist(core_set, "files_core")

    labels = {}
    speakers = {}
    intensities = {}
    for path in paths:
        spk, emo, intensity, _ = path.stem.split("_")
        labels[path.stem] = emo
        speakers[path.stem] = spk
        intensities[path.stem] = intensity

    write_annotations(labels, "label")
    write_annotations(speakers, "speaker")
    write_annotations(intensities, "intensity")


if __name__ == "__main__":
    main()



"""Scikit-learn implementations of various models.

.. autosummary::
    :toctree:

    get_sk_model
    get_sk_model_fn
    SkModelConfig


Models
------
.. autosummary::
    :toctree:

    svm
    mtl
"""

from typing import Callable

from sklearn.base import BaseEstimator
from sklearn.ensemble import RandomForestClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.neighbors import KNeighborsClassifier
from sklearn.neural_network import MLPClassifier

from ._base import ERTKSkModel, SkModelConfig, SkWrapperConfig, SkWrapperModel
from .mtl import MTFLClassifier, MultiDimensionalCC
from .svm import PrecomputedSVC

__all__ = [
    "ERTKSkModel",
    "SkModelConfig",
    "SkWrapperConfig",
    "SkWrapperModel",
    "MTFLClassifier",
    "MultiDimensionalCC",
    "PrecomputedSVC",
    "get_sk_model",
    "get_sk_model_fn",
]

CLASSIFIER_MAP = {
    "svm": PrecomputedSVC,
    "rf": RandomForestClassifier,
    "knn": KNeighborsClassifier,
    "lr": LogisticRegression,
    "mlp": MLPClassifier,
}


def get_sk_model_fn(name: str) -> Callable[..., BaseEstimator]:
    """Get a scikit-learn model class.

    Parameters
    ----------
    name: str
        The model name.

    Returns
    -------
    model_fn: callable
        A method that takes arguments and returns a Model instance.
    """
    return CLASSIFIER_MAP[name]


def get_sk_model(name: str, **kwargs) -> BaseEstimator:
    """Get a scikit-learn model by name.

    Parameters
    ----------
    name: str
        The model name.

    Returns
    -------
    model: BaseEstimator
        The unfit scikit-learn classifier.
    """
    return get_sk_model_fn(name)(**kwargs)



import re
from collections import defaultdict
from dataclasses import dataclass, field
from pathlib import Path
from typing import Dict, List, Optional, Tuple
from xml.etree import ElementTree as ET

import click
import librosa
import numpy as np
import pandas as pd
import soundfile


@dataclass
class SessionInfo:
    session_id: int
    character: str
    operator_audio_path: Optional[Path] = None
    words_operator_path: Optional[Path] = None
    user_audio_path: Optional[Path] = None
    words_user_path: Optional[Path] = None
    transcript_path: Optional[Path] = None
    annotations: Dict[str, Dict[int, Path]] = field(
        default_factory=lambda: defaultdict(dict)
    )


@click.command()
@click.argument(
    "input_dir", type=click.Path(exists=True, file_okay=False, path_type=Path)
)
@click.option("--resample/--noresample", default=True)
def main(input_dir: Path, resample: bool):
    """Processes all sessions in the SEMAINE corpus and puts combined
    audio/transcripts/ratings into the OUTPUT directory.
    """
    recordings: Dict[str, List[SessionInfo]] = defaultdict(list)
    for session_dir in (input_dir / "Sessions").glob("*"):
        session_file = session_dir / "session.xml"
        session_xml = ET.parse(session_file).getroot()

        feeltraces = []
        session_info = SessionInfo(
            int(session_dir.stem), session_xml.attrib["character"]
        )
        for path in session_dir.glob("*"):
            if re.search(r"wordLevel_alignedTranscript.*_user", path.name):
                session_info.words_user_path = path
            elif re.search(r"wordLevel_alignedTranscript.*_operator", path.name):
                session_info.words_operator_path = path
            elif re.search(r"^alignedTranscript_[0-9]+.*\.txt", path.name):
                session_info.transcript_path = path
            elif re.search(r"Operator HeadMounted.*\.wav", path.name):
                session_info.operator_audio_path = path
            elif re.search(r"User HeadMounted.*\.wav", path.name):
                session_info.user_audio_path = path
            elif re.search(r"[AR][0-9].*\.txt", path.name):
                feeltraces.append(path)
        for trace in feeltraces:
            match = re.search(
                r"[AR]([0-9]+)[RS]([0-9]+)TUC(Ob|Po|Pr|Sp)2?D([AEPV])\.txt", trace.name
            )
            if match:
                rater = int(match.group(1))
                emotion = {
                    "A": "Activation",
                    "E": "Expectation",
                    "P": "Power",
                    "V": "Valence",
                }[match.group(4)]
                session_info.annotations[emotion][rater] = trace
        recordings[session_xml.attrib["recording"]].append(session_info)

    df = pd.DataFrame(columns=["Activation", "Expectation", "Power", "Valence"])
    for recording in sorted(recordings, key=lambda x: int(x)):
        duration = 0
        concat_user_audio: List[np.ndarray] = []
        concat_operator_audio: List[np.ndarray] = []
        concat_words_user = ""
        concat_words_operator = ""
        concat_transcript = ""
        emotion_data = {}
        sessions = sorted(recordings[recording], key=lambda x: x.session_id)
        print(f"Recording {recording}:")
        for session in sessions:
            duration = int(1000 * sum(len(x) for x in concat_operator_audio) / 16000)
            audio, _ = librosa.load(session.user_audio_path, sr=16000)
            concat_user_audio.append(audio)
            audio, _ = librosa.load(session.operator_audio_path, sr=16000)
            concat_operator_audio.append(audio)

            if (
                session.character.lower() in ["beginning", "end", "forbidden"]
                or not session.words_user_path
            ):
                print(
                    f"\tSession {session.session_id} ({recording}/{session.character}) "
                    "skipped"
                )
                if not session.words_user_path:
                    print("\t(Does not have user words.)")
                print(f"\t{duration}")
                print()
                continue

            print(f"\tSession {session.session_id}:")

            for emotion in session.annotations:
                rater_data = []
                for rater in session.annotations[emotion]:
                    trace = session.annotations[emotion][rater]
                    rater_data.append(np.fromfile(trace, sep=" ").reshape((-1, 2)))

                # Truncate time to size of smallest array across raters
                smallest = np.argmin([x.shape[0] for x in rater_data])
                max_size = np.max([x.shape[0] for x in rater_data])
                num_indices = rater_data[smallest].shape[0]
                for i in range(len(rater_data)):
                    rater_data[i] = rater_data[i][:num_indices, :]

                min_size = num_indices
                print("\t", emotion, len(rater_data), rater_data[0].shape)
                if (max_size - min_size) / min_size > 0.01:
                    print(
                        "\t\t WARNING: Rating length difference: ", max_size - min_size
                    )
                mean_values = np.stack([x[:, 1] for x in rater_data], axis=0).mean(0)
                mean_raters = np.stack([rater_data[0][:, 0], mean_values], axis=1)
                mean_raters[:, 0] += duration / 1000

                if emotion not in emotion_data:
                    emotion_data[emotion] = mean_raters
                else:
                    emotion_data[emotion] = np.concatenate(
                        [emotion_data[emotion], mean_raters], axis=0
                    )

            # Truncate time to size of smallest array across emotions
            smallest_emotion = min(emotion_data, key=lambda e: emotion_data[e].shape[0])
            max_size = max([emotion_data[e].shape[0] for e in emotion_data])
            num_indices = emotion_data[smallest_emotion].shape[0]
            for e in emotion_data:
                emotion_data[e] = emotion_data[e][:num_indices, :]

            min_size = num_indices
            if max_size - min_size > 2:
                print(
                    "\t WARNING: Annotation length difference: "
                    f"{max_size - min_size}"
                )

            if session.transcript_path:
                with open(session.transcript_path) as fid:
                    concat_transcript += fid.read()
            else:
                print("\tMissing transcript")

            with open(session.words_user_path) as fid:
                for line in fid:
                    m = re.search(r'([0-9]+) ([0-9]+) ([A-Z\'"?.<>]+)', line)
                    if m:
                        line = "{} {} {}\n".format(
                            int(m.group(1)) + duration,
                            int(m.group(2)) + duration,
                            m.group(3),
                        )
                    concat_words_user += line

            if session.words_operator_path:
                with open(session.words_operator_path) as fid:
                    for line in fid:
                        m = re.search(r'([0-9]+) ([0-9]+) ([A-Z\'"?.<>]+)', line)
                        if m:
                            line = "{} {} {}\n".format(
                                int(m.group(1)) + duration,
                                int(m.group(2)) + duration,
                                m.group(3),
                            )
                        concat_words_operator += line
            else:
                print("\tMissing operator words")
            print()

        if len(emotion_data) < 4:
            print(f"Recording {recording} has incomplete emotional annotations")
            print()
            continue

        times = emotion_data["Valence"][:, 0]
        emotion_data = {e: x[:, 1] for e, x in emotion_data.items()}

        concat_user_audio = np.concatenate(concat_user_audio)
        concat_operator_audio = np.concatenate(concat_operator_audio)

        output_dir = Path("combined") / recording
        output_dir.mkdir(exist_ok=True, parents=True)
        soundfile.write(output_dir / "user.wav", concat_user_audio, samplerate=16000)
        soundfile.write(
            output_dir / "operator.wav", concat_operator_audio, samplerate=16000
        )
        with open(output_dir / "transcript.txt", "w") as fid:
            fid.write(concat_transcript)
        with open(output_dir / "user.txt", "w") as fid:
            fid.write(concat_words_user)
        with open(output_dir / "operator.txt", "w") as fid:
            fid.write(concat_words_operator)

        emotions = pd.DataFrame(emotion_data, index=pd.Index(times, name="Time"))
        emotions.to_csv(output_dir / "emotions.csv")

        user_turns: Dict[int, List[Tuple[int, int, str]]] = {}
        operator_turns: Dict[int, List[Tuple[int, int, str]]] = {}
        for d, s in [
            (user_turns, concat_words_user),
            (operator_turns, concat_words_operator),
        ]:
            turn = 0
            for line in s.split("\n"):
                if line.startswith("---"):
                    m = re.search(r"---recording.*turn ([0-9]+)---", line)
                    turn = int(m.group(1))
                else:
                    m = re.search(r"([0-9]+) ([0-9]+) <?([A-Z\'?!]+)>?", line)
                    if m:
                        start = int(m.group(1)) * 16
                        end = int(m.group(2)) * 16
                        word = m.group(3)
                        if turn not in d:
                            d[turn] = []
                        d[turn].append((start, end, word))

        for d, p, audio in [
            (user_turns, "u", concat_user_audio),
            (operator_turns, "o", concat_operator_audio),
        ]:
            out_dir = output_dir / "turns"
            out_dir.mkdir(exist_ok=True)
            for turn, words in sorted(d.items()):
                start = words[0][0]
                end = words[-1][1]

                name = f"{int(recording):02d}_{p}_{turn:03d}"
                filename = name + ".wav"
                soundfile.write(out_dir / filename, audio[start:end], samplerate=16000)

                if p == "u":
                    start_idx, end_idx = np.searchsorted(
                        emotions.index, [start / 16000, end / 16000]
                    )
                    if start_idx != end_idx:
                        mean_emotions = emotions.iloc[start_idx:end_idx, :].mean()
                        df.loc[name, :] = mean_emotions

    df.index.name = "Name"
    for c in df.columns:
        df[c].to_csv(f"{c.lower()}.csv")


if __name__ == "__main__":
    main()



"""
PyTorch backend
===============

This package contains the PyTorch backend for the ERTK.

.. autosummary::
    :toctree: generated/

    models
    train
    dataset
    classification
    utils
"""

from .models import get_pt_model, get_pt_model_fn



"""Process the raw SAVEE dataset.

This assumes the file structure from the original compressed file:
/.../
    DC/
        *.wav
    ...
"""

import re
import shutil
from pathlib import Path

import click

from ertk.dataset import resample_audio, write_annotations, write_filelist

REGEX = re.compile(r"^(DC|JE|JK|KL)([a-z][a-z]?)[0-9][0-9]$")

emotion_map = {
    "a": "anger",
    "d": "disgust",
    "f": "fear",
    "h": "happiness",
    "n": "neutral",
    "sa": "sadness",
    "su": "surprise",
}


@click.command()
@click.argument(
    "input_dir", type=click.Path(exists=True, file_okay=False, path_type=Path)
)
@click.option("--resample/--noresample", default=True)
def main(input_dir: Path, resample: bool):
    """Process the SAVEE dataset at location INPUT_DIR and resample
    audio to 16 kHz 16-bit WAV audio.
    """

    resample_dir = Path("resampled")
    if resample:
        for sp in ["DC", "JE", "JK", "KL"]:
            paths = list(input_dir.glob(sp + "/*.wav"))
            resample_audio(paths, resample_dir / sp)
        for sp in ["DC", "JE", "JK", "KL"]:
            for f in (resample_dir / sp).glob("*.wav"):
                shutil.move(f, resample_dir / (sp + f.name))
            (resample_dir / sp).rmdir()

    paths = list(resample_dir.glob("*.wav"))
    write_filelist(paths, "files_all")
    write_annotations(
        {p.stem: emotion_map[REGEX.match(p.stem).group(2)] for p in paths},
        "label",
    )
    write_annotations({p.stem: p.stem[:2] for p in paths}, "speaker")
    write_annotations({p.stem: "en" for p in paths}, "language")
    write_annotations({p.stem: "gb" for p in paths}, "country")
    write_annotations({p.stem: "M" for p in paths}, "gender")


if __name__ == "__main__":
    main()



import os
from pathlib import Path
from typing import List

import numpy as np
import pytest
from dotenv import load_dotenv

from ertk.preprocessing import (
    audioset,
    encodec,
    fairseq,
    huggingface,
    keras_apps,
    opensmile,
    openxbow,
    phonemize,
    resample,
    spectrogram,
    speechbrain,
    vad_trim,
)

test_data_dir = Path(__file__).parent / "../test_data"
files = list(test_data_dir.glob("resampled/*.wav"))


@pytest.fixture(scope="module")
def audio() -> List[np.ndarray]:
    import librosa

    return [librosa.load(x, sr=16000, mono=True)[0] for x in files]


@pytest.fixture(scope="module")
def transcript() -> np.ndarray:
    import pandas as pd

    df = pd.read_csv(test_data_dir / "transcript.csv", index_col=0)
    return df.to_numpy()


load_dotenv()
AUDIOSET_DIR = os.environ["AUDIOSET_DIR"]
FAIRSEQ_DIR = os.environ["FAIRSEQ_DIR"]


class TestAudioset:
    @pytest.mark.filterwarnings("ignore::UserWarning")
    def test_vggish(self, audio):
        config = audioset.VGGishExtractorConfig(
            model_dir=f"{AUDIOSET_DIR}/vggish", postprocess=True
        )
        ext = audioset.VGGishExtractor(config)
        feats = list(ext.process_all(audio, batch_size=32, sr=16000))
        assert len(feats) == len(files)
        assert ext.dim == 128
        assert all(x.shape == (ext.dim,) for x in feats)
        assert not ext.is_sequence

    @pytest.mark.filterwarnings("ignore::UserWarning")
    def test_yamnet(self, audio):
        config = audioset.YamnetExtractorConfig(model_dir=f"{AUDIOSET_DIR}/yamnet")
        ext = audioset.YamnetExtractor(config)
        feats = list(ext.process_all(audio, batch_size=32, sr=16000))
        assert len(feats) == len(files)
        assert ext.dim == 1024
        assert not ext.is_sequence
        assert all(x.shape == (ext.dim,) for x in feats)


class TestEncodec:
    @pytest.mark.parametrize(
        ["model", "dim"],
        [(encodec.Model.ENCODEC_48kHz, 2048), (encodec.Model.ENCODEC_24kHz, 4096)],
    )
    @pytest.mark.parametrize("aggregate", [encodec.Agg.MEAN, encodec.Agg.MAX])
    def test_encodec(self, model, aggregate, dim, audio):
        config = encodec.EncodecExtractorConfig(model=model, aggregate=aggregate)
        ext = encodec.EncodecExtractor(config)
        feats = list(ext.process_all(audio, batch_size=1, sr=16000))
        assert len(feats) == len(files)
        assert ext.dim == dim
        assert not ext.is_sequence
        assert all(x.shape == (ext.dim,) for x in feats)

    def test_encodec_vq(self, audio):
        config = encodec.EncodecExtractorConfig(
            model=encodec.Model.ENCODEC_48kHz, vq_ids=True, vq_ids_as_string=False
        )
        ext = encodec.EncodecExtractor(config)
        feats = list(ext.process_all(audio, batch_size=1, sr=16000))
        assert len(feats) == len(files)
        assert ext.dim == 16
        assert ext.is_sequence
        assert all(x.shape[1] == ext.dim for x in feats)

    def test_encodec_vq_str(self, audio):
        config = encodec.EncodecExtractorConfig(
            model=encodec.Model.ENCODEC_48kHz, vq_ids=True, vq_ids_as_string=True
        )
        ext = encodec.EncodecExtractor(config)
        feats = list(ext.process_all(audio, batch_size=1, sr=16000))
        assert len(feats) == len(files)
        assert ext.dim == 1
        assert ext.is_sequence
        assert all(x.shape[1] == ext.dim for x in feats)


class TestFairseq:
    @pytest.mark.parametrize("layer", ["context", "encoder"])
    @pytest.mark.parametrize(
        "checkpoint",
        [
            f"{FAIRSEQ_DIR}/wav2vec/wav2vec_large.pt",
            f"{FAIRSEQ_DIR}/wav2vec/vq-wav2vec.pt",
        ],
    )
    def test_fairseq_w2v(self, layer, checkpoint, audio):
        config = fairseq.FairseqExtractorConfig(
            model_type="wav2vec", checkpoint=checkpoint, layer=layer
        )
        ext = fairseq.FairseqExtractor(config)
        feats = list(ext.process_all(audio, batch_size=1, sr=16000))
        assert len(feats) == len(files)
        assert ext.dim == 512
        assert not ext.is_sequence
        assert all(x.shape == (ext.dim,) for x in feats)

    @pytest.mark.parametrize(
        ["model_type", "checkpoint", "layer", "dim"],
        [
            ("wav2vec2", f"{FAIRSEQ_DIR}/wav2vec2/libri960_big.pt", "context", 1024),
            ("wav2vec2", f"{FAIRSEQ_DIR}/wav2vec2/wav2vec_small.pt", "context", 768),
            ("wav2vec2", f"{FAIRSEQ_DIR}/wav2vec2/xlsr_53_56k.pt", "context", 1024),
            ("data2vec", f"{FAIRSEQ_DIR}/data2vec/audio_base_ls.pt", "context", 768),
            ("hubert", f"{FAIRSEQ_DIR}/hubert/hubert_base_ls960.pt", "context", 768),
            ("wav2vec2", f"{FAIRSEQ_DIR}/wav2vec2/libri960_big.pt", "encoder", 512),
            ("wav2vec2", f"{FAIRSEQ_DIR}/wav2vec2/wav2vec_small.pt", "encoder", 512),
            ("wav2vec2", f"{FAIRSEQ_DIR}/wav2vec2/xlsr_53_56k.pt", "encoder", 512),
            ("data2vec", f"{FAIRSEQ_DIR}/data2vec/audio_base_ls.pt", "encoder", 512),
            ("hubert", f"{FAIRSEQ_DIR}/hubert/hubert_base_ls960.pt", "encoder", 512),
        ],
    )
    def test_fairseq_w2v2(self, model_type, checkpoint, layer, dim, audio):
        config = fairseq.FairseqExtractorConfig(
            model_type=model_type, checkpoint=checkpoint, layer=layer
        )
        ext = fairseq.FairseqExtractor(config)
        feats = list(ext.process_all(audio, batch_size=1, sr=16000))
        assert len(feats) == len(files)
        assert ext.dim == dim
        assert not ext.is_sequence
        assert all(x.shape == (ext.dim,) for x in feats)

    def test_fairseq_hubert_vq(self, audio):
        config = fairseq.FairseqExtractorConfig(
            model_type="hubert",
            checkpoint=f"{FAIRSEQ_DIR}/hubert/hubert_base_ls960.pt",
            layer="context",
            aggregate=fairseq.Agg.NONE,
            vq_path=f"{FAIRSEQ_DIR}/textless_nlp/gslm/hubert/km200/km.bin",
            vq_ids=True,
            vq_ids_as_string=False,
        )
        ext = fairseq.FairseqExtractor(config)
        feats = list(ext.process_all(audio, batch_size=1, sr=16000))
        assert len(feats) == len(files)
        assert ext.is_sequence
        assert ext.dim == 1
        assert all(x.shape[1] == ext.dim for x in feats)


class TestHuggingFace:
    def test_hf_w2v2(self, audio):
        config = huggingface.HuggingFaceExtractorConfig(
            model="facebook/wav2vec2-base-960h", task="EMBEDDINGS", layer="context"
        )
        ext = huggingface.HuggingFaceExtractor(config)
        feats = list(ext.process_all(audio, batch_size=1, sr=16000))
        assert len(feats) == len(audio)
        assert ext.dim == 768
        assert not ext.is_sequence
        assert all(x.shape == (ext.dim,) for x in feats)


class TestKerasApps:
    @pytest.mark.parametrize("model", ["vgg16", "vgg19", "resnet50", "inception_v3"])
    def test_keras_apps(self, model, audio):
        config = keras_apps.KerasAppsExtractorConfig(model=model)
        ext = keras_apps.KerasAppsExtractor(config)
        feats = list(ext.process_all(audio, batch_size=32, sr=16000))
        assert len(feats) == len(files)
        assert not ext.is_sequence
        assert all(x.shape == (ext.dim,) for x in feats)


class TestOpenSMILE:
    def test_eGeMAPS(self, audio):
        config = opensmile.OpenSMILEExtractorConfig(opensmile_config="eGeMAPS")
        ext = opensmile.OpenSMILEExtractor(config)
        feats = list(ext.process_all(audio, batch_size=1, sr=16000))
        assert len(feats) == len(files)
        assert ext.dim == 88
        assert not ext.is_sequence
        assert all(x.shape == (1, ext.dim) for x in feats)

    def test_eGeMAPS_lld(self, audio):
        config = opensmile.OpenSMILEExtractorConfig(
            opensmile_config="eGeMAPS", levels=["lld"]
        )
        ext = opensmile.OpenSMILEExtractor(config)
        feats = list(ext.process_all(audio, batch_size=1, sr=16000))
        assert len(feats) == len(files)
        assert ext.dim == 23
        assert ext.is_sequence
        assert all(x.shape[1] == ext.dim for x in feats)


class TestOpenXBOW:
    def test_openxbow(self):
        config = openxbow.OpenXBOWExtractorConfig(
            xbowargs=["-a=10", "-size=100", "-norm=1"]
        )
        xs = np.random.randn(10, 100, 50)
        ext = openxbow.OpenXBOWExtractor(config)
        feats = list(ext.process_all(xs, batch_size=-1, sr=16000))
        assert len(feats) == 10
        assert ext.dim == 100
        assert not ext.is_sequence
        assert all(x.shape == (ext.dim,) for x in feats)


class TestPhonemize:
    def test_phonemize(self, transcript):
        config = phonemize.PhonemizeConfig(language="en-us", backend="festival")
        ext = phonemize.Phonemizer(config)
        feats = list(ext.process_all(transcript, batch_size=1, sr=16000))
        assert len(feats) == len(files)
        assert ext.dim == 1
        assert not ext.is_sequence
        assert all(x.shape == (ext.dim,) for x in feats)

    def test_phonemize_batch(self, transcript):
        config = phonemize.PhonemizeConfig(language="en-us", backend="festival")
        ext = phonemize.Phonemizer(config)
        feats = list(ext.process_all(transcript, batch_size=-1, sr=16000))
        assert len(feats) == len(files)
        assert ext.dim == 1
        assert not ext.is_sequence
        assert all(x.shape == (ext.dim,) for x in feats)


class TestResample:
    def test_resample(self, audio):
        config = resample.ResampleConfig(sample_rate=8000)
        ext = resample.Resampler(config)
        feats = list(ext.process_all(audio, batch_size=1, sr=16000))
        assert len(feats) == len(files)


class TestSpectrogram:
    def test_spectrogram(self, audio):
        config = spectrogram.SpectrogramExtractorConfig(
            window_size=0.025, window_shift=0.01, n_mels=80, fmin=0, fmax=8000
        )
        ext = spectrogram.SpectrogramExtractor(config)
        feats = list(ext.process_all(audio, batch_size=1, sr=16000))
        assert len(feats) == len(files)
        assert ext.dim == 80
        assert ext.is_sequence
        assert all(x.shape[1] == ext.dim for x in feats)


class TestSpeechBrain:
    def test_speechbrain(self, audio):
        config = speechbrain.SpeechBrainExtractorConfig(
            model="speechbrain/asr-transformer-transformerlm-librispeech",
            task=speechbrain.Task.ASR,
        )
        ext = speechbrain.SpeechBrainExtractor(config)
        feats = list(ext.process_all(audio, batch_size=1, sr=16000))
        assert len(feats) == len(files)
        assert ext.dim == 1
        assert not ext.is_sequence
        assert all(x.shape == (ext.dim,) for x in feats)


class TestVADTrim:
    def test_vad_trim(self, audio):
        config = vad_trim.VADTrimmerConfig()
        ext = vad_trim.VADTrimmer(config)
        feats = list(ext.process_all(audio, batch_size=1, sr=16000))
        assert len(feats) == len(files)



"""Audio resampling using resampy."""

from dataclasses import dataclass
from typing import List

import numpy as np
import resampy

from ertk.config import ERTKConfig

from ._base import AudioClipProcessor

__all__ = ["ResampleConfig", "Resampler"]


@dataclass
class ResampleConfig(ERTKConfig):
    """Resample configuration."""

    sample_rate: int = 16000
    """Target sample rate."""
    filter: str = "kaiser_fast"
    """Resampling filter."""


class Resampler(AudioClipProcessor, fname="resample", config=ResampleConfig):
    """Audio resampler using resampy."""

    config: ResampleConfig

    def process_instance(self, x: np.ndarray, **kwargs) -> np.ndarray:
        return resampy.resample(
            x,
            kwargs.pop("sr"),
            self.config.sample_rate,
            filter=self.config.filter,
            parallel=True,
        )

    @property
    def feature_names(self) -> List[str]:
        return ["pcm"]



"""TensorFlow training configuration."""

from dataclasses import dataclass
from typing import Optional

from ertk.config import ERTKConfig

__all__ = ["TFLoggingConfig", "TFTrainConfig"]


@dataclass
class TFLoggingConfig(ERTKConfig):
    log_dir: Optional[str] = None
    name: str = ""
    version: str = ""
    tensorboard: bool = True
    csv: bool = True


@dataclass
class TFTrainConfig(ERTKConfig):
    batch_size: int = 32
    logging: TFLoggingConfig = TFLoggingConfig()
    n_gpus: int = 1
    epochs: int = 50
    data_fn: Optional[str] = None



import pytest

from ertk.utils.generic import batch_iterable, filter_kwargs, itmap, ordered_intersect


@pytest.mark.parametrize(
    ["kwargs", "expected"],
    [
        ({"a": 1, "b": 2, "c": 3, "d": 4}, {"a": 1, "b": 2, "c": 3}),
        ({"a": 1, "b": 2, "d": 4}, {"a": 1, "b": 2}),
        ({"d": 4, "e": 5}, {}),
    ],
)
def test_filter_kwargs(kwargs, expected):
    def f(a, b, c):
        pass

    filtered = filter_kwargs(kwargs, f)
    assert filtered == expected


@pytest.mark.parametrize(
    "kwargs",
    [{"a": 1, "b": 2, "c": 3, "d": 4}, {"a": 1, "b": 2, "d": 4}, {"d": 4, "e": 5}],
)
def test_filter_kwargs_posonly(kwargs):
    filtered = filter_kwargs(kwargs, sum)
    assert filtered == {}


@pytest.mark.parametrize(
    "kwargs",
    [{"a": 1, "b": 2, "c": 3, "d": 4}, {"a": 1, "b": 2, "d": 4}, {"d": 4, "e": 5}],
)
def test_filter_kwargs_kwargs(kwargs):
    def f(*args, **kwargs):
        pass

    filtered = filter_kwargs(kwargs, f)
    assert filtered == {}


def test_itmap():
    f = itmap(lambda x: x + 1)
    assert f(1) == 2
    assert f([1, 2]) == [2, 3]
    assert type(f([1, 2])) is list
    assert f((1, 2)) == (2, 3)
    assert type(f((1, 2))) is tuple


@pytest.mark.parametrize(
    ["a", "b", "expected"],
    [
        ([5, 2, 5, 6], {5, 2}, [5, 2, 5]),
        (range(10), {5, 2}, [2, 5]),
    ],
)
def test_ordered_intersect(a, b, expected):
    assert ordered_intersect(a, b) == expected


def test_batch_iterable():
    iterable = [0] * 100
    batches = list(batch_iterable(iterable, batch_size=10, return_last=True))
    assert len(batches) == 10
    assert len(batches[0]) == 10
    assert len(batches[-1]) == 10

    batches = list(batch_iterable(iterable, batch_size=12, return_last=True))
    assert len(batches) == 9
    assert len(batches[0]) == 12
    assert len(batches[-1]) == 4

    batches = list(batch_iterable(iterable, batch_size=12, return_last=False))
    assert len(batches) == 8
    assert len(batches[0]) == 12
    assert len(batches[-1]) == 12



"""Encodec feature extractor."""

from dataclasses import dataclass
from enum import Enum
from typing import List

import numpy as np
import torch

from ertk.config import ERTKConfig
from ertk.utils import is_mono_audio

from ._base import AudioClipProcessor, FeatureExtractor

__all__ = ["EncodecExtractorConfig", "EncodecExtractor"]


class Agg(Enum):
    MEAN = "mean"
    MAX = "max"
    NONE = "none"


class Model(Enum):
    ENCODEC_24kHz = "24kHz"
    ENCODEC_48kHz = "48kHz"


@dataclass
class EncodecExtractorConfig(ERTKConfig):
    """Configuration for the Encodec feature extractor."""

    model: Model = Model.ENCODEC_48kHz
    """The Encodec model to use."""
    aggregate: Agg = Agg.MEAN
    """The aggregation method to use."""
    device: str = "cuda"
    """The device to use for inference."""
    vq_ids: bool = False
    """Whether to return the VQ-VAE IDs."""
    vq_ids_as_string: bool = True
    """Whether to return the VQ-VAE IDs as a string."""
    max_input_len: int = 1500000
    """The maximum input length for the model."""


class EncodecExtractor(
    FeatureExtractor, AudioClipProcessor, fname="encodec", config=EncodecExtractorConfig
):
    """Encodec feature extractor."""

    config: EncodecExtractorConfig

    def __init__(self, config: EncodecExtractorConfig) -> None:
        super().__init__(config)

        from encodec import EncodecModel
        from encodec.utils import convert_audio

        self.convert_audio = convert_audio

        print(f"Loading {config.model.value} model")
        if config.model == Model.ENCODEC_48kHz:
            model = EncodecModel.encodec_model_48khz()
        else:
            model = EncodecModel.encodec_model_24khz()
        torch.set_grad_enabled(False)
        self.model = model.to(device=config.device).eval()

    def process_instance(self, x: np.ndarray, **kwargs) -> np.ndarray:
        sr = kwargs.pop("sr")

        x = np.squeeze(x)
        if not is_mono_audio(x):
            raise ValueError("Audio must be mono")
        x = x[: self.config.max_input_len]
        audio = torch.as_tensor(x).unsqueeze(0)
        audio = self.convert_audio(
            audio, sr, self.model.sample_rate, self.model.channels
        ).to(device=self.config.device)
        frames = self.model.encode(audio.unsqueeze(0))
        codes = torch.cat([x[0] for x in frames], 2)
        if self.config.vq_ids:
            # (steps, n_q)
            feats = codes[0].cpu().transpose(0, 1).numpy()
            if self.config.vq_ids_as_string:
                feats = np.array([" ".join(map(str, x)) for x in feats])
                feats = np.expand_dims(feats, -1)
            return feats

        clusters = self.model.quantizer.decode(codes).cpu()
        n_q, dim, n_frames = clusters.shape
        if self.config.aggregate == Agg.MEAN:
            clusters = clusters.mean(2).reshape(n_q * dim).numpy()
        elif self.config.aggregate == Agg.MAX:
            clusters = clusters.max(2)[0].reshape(n_q * dim).numpy()
        elif self.config.aggregate == Agg.NONE:
            clusters = clusters.reshape(n_q * dim, n_frames).transpose(0, 1).numpy()
        else:
            raise ValueError(f"Invalid aggregation {self.config.aggregate}")
        return clusters

    @property
    def dim(self) -> int:
        if self.config.vq_ids:
            if self.config.vq_ids_as_string:
                return 1
            return self.model.quantizer.n_q
        return self.model.quantizer.n_q * self.model.quantizer.dimension

    @property
    def is_sequence(self) -> bool:
        return self.config.vq_ids or self.config.aggregate == Agg.NONE

    @property
    def feature_names(self) -> List[str]:
        if self.config.vq_ids:
            if self.config.vq_ids_as_string:
                return ["encodec_vq_ids"]
            return [f"encodec_vq_{i}" for i in range(self.dim)]
        return [f"encodec_{i}" for i in range(self.dim)]



"""
Utility functions
=================

This module contains miscellaneous utility functions.


Array utilities
---------------
.. autosummary::
    :toctree: generated/

    batch_arrays_by_length
    clip_arrays
    flat_to_inst
    frame_array
    frame_arrays
    inst_to_flat
    is_mono_audio
    make_array_array
    pad_arrays
    shuffle_multiple
    transpose_time


Generic utilities
-----------------
.. autosummary::
    :toctree: generated/

    batch_iterable
    filter_kwargs
    itmap
    ordered_intersect
    ordered_subsets
    subsets


Miscellaneous utilities
-----------------------
.. autosummary::
    :toctree: generated/

    TqdmMultiprocessing
    TqdmParallel


Type definitions
----------------
.. autosummary::
    :toctree: generated/

    PathOrStr
    ScoreFunction
"""

from .array import (
    batch_arrays_by_length,
    clip_arrays,
    flat_to_inst,
    frame_array,
    frame_arrays,
    inst_to_flat,
    is_mono_audio,
    make_array_array,
    pad_arrays,
    shuffle_multiple,
    transpose_time,
)
from .generic import (
    batch_iterable,
    filter_kwargs,
    itmap,
    ordered_intersect,
    ordered_subsets,
    subsets,
)
from .misc import TqdmMultiprocessing, TqdmParallel
from .types import PathOrStr, ScoreFunction



from pathlib import Path

import click
import netCDF4
import numpy as np

from ertk.dataset import write_features


@click.command()
@click.argument("input", type=click.Path(exists=True, dir_okay=False, path_type=Path))
@click.argument("corpus", type=str)
def main(input: Path, corpus: str):
    """Converts auDeep dataset INPUT to our format. Also sets corpus
    attribute to CORPUS.
    """

    dataset = netCDF4.Dataset(input, "a")
    names = [Path(x).stem for x in dataset.variables["filename"]]
    dataset.variables["filename"][:] = np.array(names)
    features = dataset.variables["features"]
    feature_names = [f"audeep{i + 1}" for i in range(features.shape[1])]
    write_features(
        input,
        names=names,
        features=features,
        corpus=corpus,
        feature_names=feature_names,
    )
    dataset.close()
    print(f"Wrote netCDF4 file to {input}")


if __name__ == "__main__":
    main()



"""Process the raw EMO-DB dataset.

This assumes the file structure from the original compressed file:
/.../
    wav_corpus/
        *.wav
    silb/
        ...
    ...
"""

import shutil
from pathlib import Path

import click
from tqdm import tqdm

from ertk.dataset import write_annotations, write_filelist

sentence_map = {
    "a01": "Der Lappen liegt auf dem Eisschrank.",
    "a02": "Das will sie am Mittwoch abgeben.",
    "a04": "Heute abend könnte ich es ihm sagen.",
    "a05": "Das schwarze Stück Papier befindet sich da oben neben dem Holzstück.",
    "a07": "In sieben Stunden wird es soweit sein.",
    "b01": "Was sind denn das für Tüten, die da unter dem Tisch stehen?",
    "b02": "Sie haben es gerade hochgetragen und jetzt gehen sie wieder runter.",
    "b03": "An den Wochenenden bin ich jetzt immer nach Hause gefahren und habe Agnes besucht.",  # noqa E501
    "b09": "Ich will das eben wegbringen und dann mit Karl was trinken gehen.",
    "b10": "Die wird auf dem Platz sein, wo wir sie immer hinlegen.",
}

emotion_map = {
    "W": "anger",
    "L": "boredom",
    "E": "disgust",
    "A": "fear",
    "F": "happiness",
    "T": "sadness",
    "N": "neutral",
}


@click.command()
@click.argument(
    "input_dir", type=click.Path(exists=True, file_okay=False, path_type=Path)
)
@click.option("--resample/--noresample", default=True)
def main(input_dir: Path, resample: bool):
    """Process the EMO-DB dataset at location INPUT_DIR."""
    paths = list(input_dir.glob("wav_corpus/*.wav"))

    speaker_dict = {p.stem: p.stem[:2] for p in paths}
    write_annotations(speaker_dict, "speaker")
    male_speakers = ["03", "10", "11", "12", "15"]
    gender_dict = {
        k: "M" if v in male_speakers else "F" for k, v in speaker_dict.items()
    }
    write_annotations(gender_dict, "gender")
    write_annotations({p.stem: p.stem[2:5] for p in paths}, "sentence")
    write_annotations({p.stem: sentence_map[p.stem[2:5]] for p in paths}, "transcript")
    write_annotations({p.stem: emotion_map[p.stem[5]] for p in paths}, "label")
    write_annotations({p.stem: p.stem[6] for p in paths}, "version")
    write_annotations({p.stem: "de" for p in paths}, "language")

    utt = {}
    for p in input_dir.glob("silb/*.silb"):
        with open(p, encoding="latin_1") as fid:
            words = []
            for line in fid:
                line = line.strip()
                _, word = line.split()
                if word in [".", "("]:
                    continue
                words.append(word.strip())
            utt[p.stem] = " ".join(words)
    write_annotations(utt, "syllables")

    if resample:
        resample_dir = Path("resampled")
        resample_dir.mkdir(exist_ok=True)
        for p in tqdm(paths, desc="Copying audio"):
            shutil.copyfile(p, resample_dir / p.name)
        paths = list(resample_dir.glob("*.wav"))
    write_filelist(paths, "files_all")


if __name__ == "__main__":
    main()



"""Process the raw IEMOCAP dataset.

This assumes the file structure from the original compressed file:
/.../
    Session1/
        sentences/
            wav/
                Ses01F_impro01/
                    *.wav
                ...
            ...
        ...
    ...
"""

import re
from pathlib import Path

import click
import numpy as np
import pandas as pd

from ertk.dataset import resample_audio, write_annotations, write_filelist
from ertk.stats import alpha

# [START_TIME - END_TIME] TURN_NAME EMOTION [V, A, D]
REGEX = re.compile(
    r"^\[(\d+\.\d+) - (\d+\.\d+)\]\t(\w+)\t(\w+)\t\[(\d+\.\d+), (\d+\.\d+), (\d+\.\d+)\]$"  # noqa
)
TRN_REGEX = re.compile(
    r"^(Ses0[1-5][MF]_(?:impro|script)0[1-9][ab]?(?:_\db?)?_[MF][X\d]{2}\d) \[\d+\.\d+-\d+\.\d+\]:(.*)$"  # noqa
)

emotion_map = {
    "ang": "anger",
    "hap": "happiness",
    "sad": "sadness",
    "neu": "neutral",
    "exc": "excitement",
    "dis": "disgust",
    "fru": "frustration",
    "fea": "fear",
    "sur": "surprise",
    "oth": "other",
    "xxx": "unknown",
}

ratings_map = {"excited": "excitement"}


def _clean(words: str):
    words = " ".join(re.split(r"\[\w+\]", words))
    words = " ".join(words.split("..."))
    words = " ".join(words.split(".."))
    words = " ".join(words.split("-"))
    return " ".join((x.strip() for x in words.split()))


@click.command()
@click.argument(
    "input_dir", type=click.Path(exists=True, file_okay=False, path_type=Path)
)
@click.option("--resample/--noresample", default=True)
def main(input_dir: Path, resample: bool):
    """Process the IEMOCAP dataset at location INPUT_DIR and resample
    audio to 16 kHz 16-bit WAV audio.
    """

    dimensions = {}
    labels = {}
    _cat_ratings = []
    _att_ratings = []
    for filename in input_dir.glob("Session?/dialog/EmoEvaluation/*.txt"):
        with open(filename) as fid:
            name = ""
            for line in fid:
                line = line.strip()
                match = REGEX.match(line)
                if match:
                    name = match.group(3)
                    labels[name] = match.group(4)
                    dimensions[name] = list(map(float, match.group(5, 6, 7)))
                elif line.startswith("C"):
                    # Emotion classification
                    rater, _annotations = line.strip().split(":")
                    rater = rater.strip().split("-")[1]
                    if rater[0] in "MF":
                        # M or F refer to self-evalulation
                        continue
                    *annotations, comments = _annotations.strip().split(";")
                    label = annotations[0].strip().lower()
                    _cat_ratings.append((name, rater, ratings_map.get(label, label)))
                elif line.startswith("A"):
                    # Emotion attributes
                    rater, _annotations = line.strip().split(":")
                    rater = rater.strip().split("-")[1]
                    if rater[0] in "MF":
                        # M or F refer to self-evalulation
                        continue
                    *annotations, comments = _annotations.strip().split(";")
                    val = int(annotations[0].strip().split(" ")[1])
                    act = int(annotations[1].strip().split(" ")[1])
                    try:
                        # For some reason 36 of these are missing
                        dom = int(annotations[2].strip().split("  ")[1])
                    except IndexError:
                        dom = 0
                    _att_ratings.append((name, rater, val, act, dom))

    paths = list(input_dir.glob("Session?/sentences/wav/*/*.wav"))
    if resample:
        resample_dir = Path("resampled")
        resample_audio(paths, resample_dir)
        paths = list(resample_dir.glob("*.wav"))
    write_filelist(paths, "files_all")
    write_filelist(
        [p for p in paths if labels[p.stem] in ["ang", "hap", "neu", "sad", "exc"]],
        "files_4class",
    )

    write_annotations({n: emotion_map[labels[n]] for n in labels}, "label")
    speaker_dict = {p.stem: p.stem[3:6] for p in paths}
    write_annotations(speaker_dict, "speaker")
    write_annotations({k: v[-1] for k, v in speaker_dict.items()}, "gender")
    write_annotations({k: v[:2] for k, v in speaker_dict.items()}, "session")
    write_annotations(
        {p.stem: "impro" if "impro" in p.stem else "script" for p in paths}, "sess_type"
    )
    write_annotations({p.stem: "en" for p in paths}, "language")
    write_annotations({p.stem: "us" for p in paths}, "country")

    # Aggregated dimensional annotations per utterance
    df = pd.DataFrame.from_dict(
        dimensions, orient="index", columns=["valence", "activation", "dominance"]
    )
    df.index.name = "name"
    for dim in ["valence", "activation", "dominance"]:
        df[dim].to_csv(dim + ".csv", index=True, header=True)

    dim_ratings = (
        pd.DataFrame(_att_ratings, columns=["name", "rater", "val", "act", "dom"])
        .set_index(["name", "rater"])
        .sort_index()
    )
    cat_ratings = (
        pd.DataFrame(_cat_ratings, columns=["name", "rater", "label"])
        .set_index(["name", "rater"])
        .sort_index()
    )
    ratings = cat_ratings.join(dim_ratings)
    ratings.to_csv("ratings.csv")

    ratings["label"] = ratings["label"].str.lower()
    ratings["label"] = ratings["label"].map(
        lambda x: {"excited": "excitement"}.get(x, x)
    )
    ratings["label"] = ratings["label"].astype("category")

    # There are 3 (non-self) raters per utterance. Agreement is the
    # proportion of labels which are the same. This formula only works
    # for 3 raters.
    mode_count = 4 - ratings["label"].unstack().nunique(1)
    agreement = np.mean(mode_count / 3)
    print(mode_count.value_counts().rename_axis("majority").to_frame("count"))
    print(f"Mean label agreement: {agreement:.3f}")

    # Simple way to get int matrix of labels for raters x clips
    data = ratings["label"].cat.codes.unstack() + 1
    data[data.isna()] = 0
    data = data.astype(int).to_numpy()
    print(f"Krippendorf's alpha (categorical): {alpha(data):.3f}")

    # Transcriptions
    utts = {}
    for p in input_dir.glob("Session?/dialog/transcriptions/*.txt"):
        with open(p) as fid:
            for line in map(str.strip, fid):
                match = TRN_REGEX.match(line)
                if match:
                    utts[match.group(1)] = match.group(2).strip()
    utts = {n: _clean(u) for n, u in utts.items()}
    df = pd.DataFrame.from_dict(utts, orient="index", columns=["transcript"])
    df = df.rename_axis(index="name").sort_index()
    df.to_csv("transcript.csv", header=True, index=True)


if __name__ == "__main__":
    main()



from pathlib import Path
from typing import List, Tuple

import click
import pytorch_lightning as pl
import torch
from omegaconf import OmegaConf
from torch.utils.data import DataLoader, Dataset

from ertk.pytorch import get_pt_model
from ertk.pytorch.models import PyTorchModelConfig


class RandomDataset(Dataset[Tuple[torch.Tensor, ...]]):
    def __init__(
        self, total: int, shapes: List[Tuple[int]], dtypes: List[torch.dtype]
    ) -> None:
        super().__init__()
        self.total = total
        self.data = []
        for shape, dtype in zip(shapes, dtypes):
            self.data.append(torch.zeros(shape, dtype=dtype))

    def __getitem__(self, index) -> Tuple[torch.Tensor, ...]:
        return tuple(self.data)

    def __len__(self) -> int:
        return self.total


@click.command()
@click.argument("model_name")
@click.option("--batch_size", type=int, default=32, help="Batch size.")
@click.option("--shape", default="[(1024,), ()]", help="Shapes.")
@click.option("--dtype", default="[torch.float32, torch.int64]", help="Dtypes.")
@click.option(
    "--summary_only", is_flag=True, help="Print summary but don't train on dummy data."
)
@click.option("--model_config", type=click.Path(exists=True, path_type=Path))
@click.option("--train", type=int, default=8000, help="Number of training instances.")
@click.option("--valid", type=int, default=800, help="Number of validation instances.")
@click.argument("restargs", nargs=-1)
def test_fit(
    model_name: str,
    batch_size: int,
    shape: str,
    dtype: str,
    summary_only: bool,
    model_config: Path,
    train: int,
    valid: int,
    restargs: Tuple[str],
):
    """Trains a model for 2 epochs with random data and prints timing
    information for testing purposes.
    """

    config: PyTorchModelConfig
    if model_config:
        config = OmegaConf.load(model_config)
    else:
        config = PyTorchModelConfig(
            optimiser="adam", learning_rate=0.001, loss="cross_entropy"
        )

    cli_args = OmegaConf.from_cli(list(restargs))
    if cli_args.model_config:
        config = OmegaConf.merge(config, cli_args.model_config)

    model = get_pt_model(model_name, config=config)
    print(model)
    trainable_params = sum(p.numel() for p in model.parameters() if p.requires_grad)
    total_params = sum(p.numel() for p in model.parameters())
    print("Trainable parameters: {}".format(trainable_params))
    print("Total parameters: {}".format(total_params))
    print()

    if summary_only:
        return

    shapes = eval(shape, {})
    dtypes = eval(dtype, {"torch": torch})
    train_data = RandomDataset(train, shapes, dtypes)
    valid_data = RandomDataset(valid, shapes, dtypes)
    train_dl = DataLoader(
        train_data, batch_size=batch_size, shuffle=True, pin_memory=True
    )
    valid_dl = DataLoader(
        valid_data, batch_size=batch_size, shuffle=False, pin_memory=True
    )

    trainer = pl.Trainer(
        gpus=1,
        logger=False,
        enable_progress_bar=True,
        enable_model_summary=False,
        enable_checkpointing=False,
        max_epochs=2,
    )
    trainer.fit(model, train_dataloaders=train_dl, val_dataloaders=valid_dl)
    trainer.predict(model, valid_dl)


if __name__ == "__main__":
    test_fit()
